from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.top import TopAttributes


class CodeSigningEnvironmentTemplateBaseAttributes(TopAttributes, metaclass=IterableMeta):
	__config_class__ = "Code Signing Environment Template Base"
	environment_template_update_flow_dn = Attribute('Environment Template Update Flow DN', min_version='21.4')
	flow_instance_macro = Attribute('Flow Instance Macro', min_version='20.1')
	key_use_authentication = Attribute('Key Use Authentication', min_version='20.1')
	key_use_flow_dn = Attribute('Key Use Flow DN', min_version='20.1')
	naming_pattern = Attribute('Naming Pattern', min_version='20.3')
	per_user = Attribute('Per User', min_version='20.3')
	require_justification = Attribute('Require Justification', min_version='20.1')
	user_key_import = Attribute('User Key Import', min_version='20.3')
	visible_to = Attribute('Visible To', min_version='20.1')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.top import TopAttributes


class GroupAttributes(TopAttributes, metaclass=IterableMeta):
	__config_class__ = "Group"
	assets = Attribute('Assets', min_version='20.1')
	closed_group = Attribute('Closed Group', min_version='20.1')
	foreign_security_principal = Attribute('Foreign Security Principal', min_version='19.3')
	full_name = Attribute('Full Name')
	group_membership = Attribute('Group Membership')
	member = Attribute('Member')
	owner = Attribute('Owner', min_version='20.1')
	products = Attribute('Products', min_version='20.1')
	suggested_member = Attribute('Suggested Member', min_version='20.1')
	team_member_added_by = Attribute('Team Member Added By', min_version='20.3')
	team_member_added_on = Attribute('Team Member Added On', min_version='20.3')



from typing import List, Dict, Union
from pytpp.tools.vtypes import Config, Identity
from pytpp.features.bases.feature_base import FeatureBase, feature
from pytpp.features.definitions.exceptions import UnexpectedValue
from pytpp.attributes.discovery import DiscoveryAttributes


@feature('Network Discovery')
class NetworkDiscovery(FeatureBase):
    def __init__(self, api):
        super().__init__(api=api)
        self._discovery_dn = r'\VED\Discovery'

    def create(self, name: str, hosts: List[str], default_certificate_location: 'Union[Config.Object, str]',
               attributes: dict = None, automatically_import: bool = False, blackout: Dict[str, List] = None,
               contacts: 'List[Union[Identity, str]]' = None, days_of_week: List[str] = None, days_of_month: List[str] = None,
               days_of_year: List[str] = None, description: str = None,
               exclusion_locations: 'Union[Config.Object, str]' = None, hour: int = None,
               placement_rules: 'List[Union[Config.Object, str]]' = None,
               ports: List[Union[str, int]] = None, priority: int = None, reschedule: bool = True,
               resolve_host: bool = True, utc: str = '1', get_if_already_exists: bool = True):
        """
        Args:
            name: Name of the discovery job.
            hosts: A list of hosts. If specific hosts should scan different ports, then specify by appending the
                   port to the IP address or hostname (i.e. 192.168.0.10:80). If no port is specified, then the
                   `ports` parameter will be appended to those hosts.
            default_certificate_location: :ref:`config_object` or :ref:`dn` of the folder to place results not 
                                          matching a placement rule.
            attributes: Additional attributes.
            automatically_import: If `True`, the job will terminate by placing the certificate objects, device
                                  objects, and application objects according to the placement rules.
            blackout: Period of time that TPP will not perform discovery operations. The format of this value
                      should be as follows: `[{<day>:[<hour>, <hour>, ...]}, {...}]` where `day` is the
                      zero-based index day of the week (i.e. Sunday = '0', etc.) and `hour` is the 24-hour
                      hour of the day (i.e. 1, 2, 3, ..., 23). For example:
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` as contacts.
            days_of_week: Zero-based index value of the days of the week to run the job.
            days_of_month: Day value(s) of the month to run the job.
            days_of_year: Days of the year to run the job in the format "MM/DD" where leading zeros can be ignored
                          (i.e. 1/23, 10/3).
            description: Description of the job.
            exclusion_locations: List of :ref:`config_object` or :ref:`dn` of exclusion folders.
            hour: 24-hour UTC hour format of the day (i.e. 20 = 8 PM UTC).
            placement_rules: List of :ref:`config_object` or :ref:`dn` for the placement rules. The order of the list matters
                             as the rules are prioritized accordingly.
            ports: List of ports to scan.
            priority: Priority of the job.
            reschedule: When ``True``, the job will run again on the next scheduled interval.
            resolve_host: Resolve the hostname when ``True``.
            utc: UTC offset.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the discovery job.

        """
        placement_rule_guids = None
        if placement_rules:
            placement_rule_guids = [f'{e}:{self._get_guid(pr)}' for e, pr in enumerate(placement_rules)]
        ports = ','.join(list(map(str, ports))) if ports else \
            "22,80,443-449,465,563,636,695,981-995,1311,1920,2083,2087,2096,2211,2484,2949,3268,3269,3414," \
            "4712,4843,5223,5358,6619,6679,6697,7002,8080,8222,8243,8333,8443,8878,8881,8882,9043,9090," \
            "9091,9443,18072,18080-18085,18090-18094,28080"
        address_range = [f'{h}:{ports}' if ':' not in h else h for h in hosts]

        if blackout:
            blackout = [f'{k}:{",".join(v)}' for k, v in blackout.items()]

        attributes = attributes or {}
        attributes.update({
            DiscoveryAttributes.address_range          : address_range,
            DiscoveryAttributes.automatically_import   : "1" if automatically_import else "0",
            DiscoveryAttributes.blackout               : blackout,
            DiscoveryAttributes.certificate_location_dn: self._get_dn(default_certificate_location)
                                                         if default_certificate_location else None,
            DiscoveryAttributes.contact                : [self._get_prefixed_universal(c) for c in contacts] if contacts else None,
            DiscoveryAttributes.description            : description,
            DiscoveryAttributes.discovery_exclusion_dn : [self._get_dn(el) for el in exclusion_locations]
                                                         if exclusion_locations else None,
            DiscoveryAttributes.placement_rule         : placement_rule_guids,
            DiscoveryAttributes.priority               : priority,
            DiscoveryAttributes.reschedule             : "1" if hour and reschedule else "0",
            DiscoveryAttributes.resolve_host           : "1" if resolve_host else "0",
            DiscoveryAttributes.utc                    : utc
        })

        if hour:
            attributes[DiscoveryAttributes.hour] = hour
            if days_of_week:
                attributes[DiscoveryAttributes.days_of_week] = days_of_week
            elif days_of_month:
                attributes[DiscoveryAttributes.days_of_month] = days_of_month
            elif days_of_year:
                attributes[DiscoveryAttributes.days_of_year] = days_of_year

        return self._config_create(
            name=name,
            parent_folder_dn=self._discovery_dn,
            config_class=DiscoveryAttributes.__config_class__,
            attributes=attributes,
            get_if_already_exists=get_if_already_exists
        )

    def delete(self, job: 'Union[Config.Object, str]'):
        """
        Deletes the discovery job.

        Args:
            job: :ref:`config_object` or :ref:`dn` of the discovery job.
        """
        job_guid = self._get_guid(job, parent_dn=self._discovery_dn)
        response = self._api.websdk.Discovery.Guid(guid=job_guid).delete()
        response.assert_valid_response()

    def get(self, name: str, raise_error_if_not_exists: bool = True):
        """
        Args:
            name: Name of the discovery job.
            raise_error_if_not_exists: Raise an exception if the discovery job does not exist.

        Returns:
            :ref:`config_object` of the discovery job.
        """
        return self._get_config_object(
            object_dn=f'{self._discovery_dn}\\{name}',
            raise_error_if_not_exists=raise_error_if_not_exists
        )

    def is_in_progress(self, job: 'Union[Config.Object, str]'):
        """
        Args:
            job: :ref:`config_object` or :ref:`dn` of the discovery job.

        Returns:
            bool: ``True`` if the job is in progress or ``False`` if it is not.
        """
        job_dn = self._get_dn(job, parent_dn=self._discovery_dn)
        response = self._api.websdk.Config.Read.post(
            object_dn=job_dn,
            attribute_name=DiscoveryAttributes.status
        )
        in_progress_states = ['Pending Execution', 'Running']
        if response.is_valid_response():
            if len(response.values) > 0:
                status = response.values[0]
                return status in in_progress_states
        return False

    def schedule(self, job: 'Union[Config.Object, str]', hour: Union[str, int], days_of_week: List[Union[int, str]] = None,
                 days_of_month: List[Union[int, str]] = None, days_of_year: List[str] = None):
        """
        Schedules an existing job.

        Args:
            job: :ref:`config_object` or :ref:`dn` of the discovery job.
            hour: 24-hour UTC hour format (i.e. 20 = 8PM UTC).
            days_of_week: Zero-based index of the days of the week (i.e. Sunday = '0').
            days_of_month: Days of the month without leading zeros.
            days_of_year: Days of the year in "MM/DD" format without leading zeros (i.e. 1/23, 10/3).
        """
        job_dn = self._get_dn(job, parent_dn=self._discovery_dn)
        hour = str(hour)
        if not isinstance(hour, list):
            hour = [hour]

        attributes = {
            DiscoveryAttributes.reschedule: "1",
            DiscoveryAttributes.hour      : hour
        }
        if days_of_week:
            attributes[DiscoveryAttributes.days_of_week] = map(str, days_of_week)
        elif days_of_month:
            attributes[DiscoveryAttributes.days_of_month] = map(str, days_of_month)
        elif days_of_year:
            attributes[DiscoveryAttributes.days_of_year] = days_of_year

        response = self._api.websdk.Config.Write.post(
            object_dn=job_dn,
            attribute_data=self._name_value_list(attributes, keep_list_values=True)
        )
        response.assert_valid_response()

    def unschedule(self, job: 'Union[Config.Object, str]'):
        """
        Removes a schedule from a job. This does not delete the job.

        Args:
            job: :ref:`config_object` or :ref:`dn` of the discovery job.
        """
        job_dn = self._get_dn(job, parent_dn=self._discovery_dn)
        for attribute_name in {
            DiscoveryAttributes.hour,
            DiscoveryAttributes.days_of_year,
            DiscoveryAttributes.days_of_month,
            DiscoveryAttributes.days_of_week,
            DiscoveryAttributes.reschedule
        }:
            self._api.websdk.Config.ClearAttribute.post(
                object_dn=job_dn,
                attribute_name=attribute_name
            ).assert_valid_response()

    def blackout_schedule(self, job: 'Union[Config.Object, str]', sunday: List[Union[str, int]] = None,
                          monday: List[Union[str, int]] = None, tuesday: List[Union[str, int]] = None,
                          wednesday: List[Union[str, int]] = None, thursday: List[Union[str, int]] = None,
                          friday: List[Union[str, int]] = None, saturday: List[Union[str, int]] = None):
        """
        Times of the week to restrict a discovery job from processing.

        Args:
            job: :ref:`config_object` or :ref:`dn` of the discovery job.
            sunday: List of hours without leading zeros to restrict processing on Sunday.
            monday: List of hours without leading zeros to restrict processing on Monday.
            tuesday: List of hours without leading zeros to restrict processing on Tuesday.
            wednesday: List of hours without leading zeros to restrict processing on Wednesday.
            thursday: List of hours without leading zeros to restrict processing on Thursday.
            friday: List of hours without leading zeros to restrict processing on Friday.
            saturday: List of hours without leading zeros to restrict processing on Saturday.
        """
        job_dn = self._get_dn(job, parent_dn=self._discovery_dn)
        blackout = []
        for e, day in enumerate([sunday, monday, tuesday, wednesday, thursday, friday, saturday]):
            if day:
                hours = ','.join(map(str, day))
                blackout.append(f'{e}:{hours}')
        attributes = {
            DiscoveryAttributes.blackout: blackout
        }
        response = self._api.websdk.Config.Write.post(
            object_dn=job_dn,
            attribute_data=self._name_value_list(attributes, keep_list_values=True)
        )
        response.assert_valid_response()

    def run_now(self, job: 'Union[Config.Object, str]', timeout: int = 60):
        """
        Runs a job despite any scheduling. This does not return until the job is processing,
        or has a *Processing* Attribute.

        Args:
            job: :ref:`config_object` or :ref:`dn` of the discovery job.
            timeout: Timeout in seconds within which the job should start.
        """
        job_dn = self._get_dn(job, parent_dn=self._discovery_dn)
        response = self._api.websdk.Config.Write.post(
            object_dn=job_dn,
            attribute_data=self._name_value_list({
                "Start Now": ['1']  # Secret config-bridge attribute name.
            })
        )
        response.assert_valid_response()

        with self._Timeout(timeout=timeout) as to:
            while not to.is_expired():
                if self.is_in_progress(job=job):
                    return

        raise UnexpectedValue(
            f'Expected the job "{job_dn}" to start progress, but it did not.'
        )

    def cancel(self, job: 'Union[Config.Object, str]'):
        """
        Cancels a currently running job.

        Args:
            job: :ref:`config_object` or :ref:`dn` of the discovery job.
        """
        job_dn = self._get_dn(job, parent_dn=self._discovery_dn)
        response = self._api.websdk.Config.WriteDn.post(
            object_dn=job_dn,
            attribute_name=DiscoveryAttributes.status,
            values=['Canceled']
        )
        response.assert_valid_response()

    def pause(self, job: 'Union[Config.Object, str]'):
        """
        Pauses a currently running job.

        Args:
            job: :ref:`config_object` or :ref:`dn` of the discovery job.
        """
        job_dn = self._get_dn(job, parent_dn=self._discovery_dn)
        response = self._api.websdk.Config.WriteDn.post(
            object_dn=job_dn,
            attribute_name=DiscoveryAttributes.status,
            values=['Paused']
        )
        response.assert_valid_response()

    def resume(self, job: 'Union[Config.Object, str]'):
        """
        Resumes a currently paused job.

        Args:
            job: :ref:`config_object` or :ref:`dn` of the discovery job.
        """
        job_dn = self._get_dn(job, parent_dn=self._discovery_dn)
        response = self._api.websdk.Config.WriteDn.post(
            object_dn=job_dn,
            attribute_name=DiscoveryAttributes.status,
            values=['Pending']
        )
        response.assert_valid_response()

    def place_results(self, job: 'Union[Config.Object, str]'):
        """
        .. warning::
            This functionality has been deprecated in TPP 21.1 and will have no effect
            from this version forward.

        Places the results of the discovery job according to the placement rules.

        Args:
            job: :ref:`config_object` or :ref:`dn` of the discovery job.
        """
        if self._is_version_compatible(maximum="20.4"):
            job_dn = self._get_dn(job, parent_dn=self._discovery_dn)
            response = self._api.websdk.Config.WriteDn.post(
                object_dn=job_dn,
                attribute_name="Import Results Now",  # Secret config-bridge attribute name.
                values=['1']
            )
            response.assert_valid_response()

    def get_all_jobs(self):
        """
        Returns:
            List of :ref:`config_object` of all network discovery jobs.
        """
        jobs = self._api.websdk.Config.FindObjectsOfClass.post(
            class_name='Discovery',
            object_dn=self._discovery_dn
        )

        return jobs.objects

    def wait_for_job_to_finish(self, job: 'Union[Config.Object, str]', check_interval: int = 5, timeout: int = 300):
        """
        Waits for the  *Status* attribute to have a value other than *Pending Execution* and *Running*
        on the discovery job. An error is raised if the timeout is exceeded.

        Args:
            job: :ref:`config_object` or :ref:`dn` of the discovery job.
            check_interval: Poll interval in seconds to validate that the job finished.
            timeout: Timeout in seconds to wait for the job to finish.
        """
        job_dn = self._get_dn(job, parent_dn=self._discovery_dn)
        with self._Timeout(timeout=timeout) as to:
            while not to.is_expired(poll=check_interval):
                if not self.is_in_progress(job=job):
                    return

        status = self._api.websdk.Config.Read.post(
            object_dn=job_dn,
            attribute_name=DiscoveryAttributes.status
        ).values[0]
        raise TimeoutError(
            f'Expected Network Discovery Job "{job_dn}" to finish within {timeout} seconds, but it is still '
            f'running. It has a status of "{status}."'
        )



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.service_module import ServiceModuleAttributes


class BulkProvisioningManagerAttributes(ServiceModuleAttributes, metaclass=IterableMeta):
	__config_class__ = "Bulk Provisioning Manager"
	retry_interval = Attribute('Retry Interval', min_version='18.3')



from typing import List 
from pytpp.api.api_base import API, APIResponse, api_response_property


class _Discovery:
    def __init__(self, api_obj):
        self._api_obj = api_obj
        self.Import = self._Import(api_obj=api_obj)

    def Guid(self, guid: str):
        return self._Guid(guid=guid, api_obj=self._api_obj)

    class _Guid(API):
        def __init__(self, guid: str, api_obj):
            super().__init__(api_obj=api_obj, url=f'/Discovery/{guid}')

        def delete(self):
            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def success(self) -> bool:
                    return self._from_json('Success')

            return _Response(response=self._delete())

    class _Import(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Discovery/Import')

        def post(self, endpoints: list, zone_name: str):
            body = {
                'zoneName': zone_name,
                'endpoints': endpoints
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def created_certificates(self) -> int:
                    return self._from_json('createdCertificates')

                @property
                @api_response_property()
                def created_instances(self) -> int:
                    return self._from_json('createdInstances')

                @property
                @api_response_property()
                def updated_certificates(self) -> int:
                    return self._from_json('updatedCertificates')

                @property
                @api_response_property()
                def updated_instances(self) -> int:
                    return self._from_json('updatedInstances')

                @property
                @api_response_property()
                def warnings(self) -> List[str]:
                    return self._from_json('warnings')

                @property
                @api_response_property()
                def zone_name(self) -> str:
                    return self._from_json('zoneName')

            return _Response(response=self._post(data=body))




from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.top import TopAttributes


class ReportTemplateAttributes(TopAttributes, metaclass=IterableMeta):
	__config_class__ = "Report Template"
	item_vault_id = Attribute('Item Vault Id')
	report_class = Attribute('Report Class')
	template_vault_id = Attribute('Template Vault Id')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.certificate_trust_store_base import CertificateTrustStoreBaseAttributes


class BlueCoatSSLVATrustStoreAttributes(CertificateTrustStoreBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "BlueCoat SSLVA Trust Store"
	create_lists = Attribute('Create Lists')
	key_store = Attribute('Key Store')



from pytpp.properties.response_objects.dataclasses import preferences


class Preferences:
    @staticmethod
    def Preference(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return preferences.Preference(
            id=response_object.get('Id'),
            universal=response_object.get('Universal'),
            product=response_object.get('Product'),
            category=response_object.get('Category'),
            name=response_object.get('Name'),
            value=response_object.get('Value'),
            locked=response_object.get('Locked'),
        )



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.top import TopAttributes


class LogNotificationAttributes(TopAttributes, metaclass=IterableMeta):
	__config_class__ = "Log Notification"
	log_channel = Attribute('Log Channel')
	rule = Attribute('Rule')



from dataclasses import dataclass
from typing import List, Union, TYPE_CHECKING
if TYPE_CHECKING:
    from pytpp.tools.vtypes import Config, Identity
from pytpp.attributes.x509_certificate import X509CertificateAttributes
from pytpp.features.bases.feature_base import FeatureBase, feature
from pytpp.features.definitions.exceptions import FeatureException, UnexpectedValue
from pytpp.features.definitions.classes import Classes
from pytpp.tools.logger import logger, LogTags


@dataclass
class DownloadedCertificate:
    certificate_data: str
    filename: str
    format: str


@feature('Certificate')
class Certificate(FeatureBase):
    def __init__(self, api):
        super().__init__(api)

    def _get(self, certificate: 'Union[Config.Object, str]'):
        # High volume concurrency can cause a 500 internal error in IIS due to a deadlock.
        # In this case the error requests the client to "rerun the transaction".
        certificate_guid = self._get_guid(certificate)

        retries = 3
        result = self._api.websdk.Certificates.Guid(certificate_guid).get()
        for retry in range(retries):
            if 'Rerun the transaction' not in result.api_response.reason:
                result.assert_valid_response()
                return result
            logger.log('Rerunning Certificates/Get transaction due to deadlock...', log_tag=LogTags.feature)
            result = self._api.websdk.Certificates.Guid(certificate_guid).get()
        # It's likely at this point that we failed to get the certificate. Let the result.assert_valid_response()
        # handle the errors. In some unknown case where the response is valid, return it just in case.
        result.assert_valid_response()
        return result

    def associate_application(self, certificate: 'Union[Config.Object, str]', applications: 'List[Union[Config.Object, str]]',
                              push_to_new: bool = False):
        """
        Associates an application object to a certificate object.

        Args:
            certificate: :ref:`config_object` or :ref:`dn` of the certificate object.
            applications: A list of :ref:`config_object` or :ref:`dn` for each application object.
            push_to_new: If ``True``, the certificate will be pushed to the application once associated.
        """
        certificate_dn = self._get_dn(certificate)
        result = self._api.websdk.Certificates.Associate.post(
            application_dn=[self._get_dn(a) for a in applications],
            certificate_dn=certificate_dn,
            push_to_new=push_to_new
        )
        if not result.success:
            raise FeatureException(f'Unable to associate the given applications to the certificate "{certificate_dn}".')

    def create(self, name: str, parent_folder: 'Union[Config.Object, str]', description: 'str' = None,
               contacts: 'List[Union[Identity.Identity, str]]' = None, approvers: 'List[Union[Identity.Identity, str]]' = None,
               management_type: 'str' = None, service_generated_csr: 'bool' = None, generate_key_on_application: 'bool' = None,
               hash_algorithm: 'str' = None, common_name: 'str' = None, organization: 'str' = None, organization_unit: 'List[str]' = None,
               city: 'str' = None, state: 'str' = None, country: 'str' = None, san_dns: 'List[str]' = None, san_email: 'List[str]' = None,
               san_upn: 'List[str]' = None, san_ip: 'List[str]' = None, san_uri: 'List[str]' = None, key_algorithm: 'str' = None,
               key_strength: 'int' = None, elliptic_curve: 'str' = None, ca_template: 'Union[Config.Object, str]' = None,
               disable_automatic_renewal: 'bool' = None, renewal_window: 'int' = None, attributes: dict = None,
               get_if_already_exists: bool = True):
        """
        .. note::
            The certificate is not automatically requested. Use :meth:`renew` to obtain a certificate.

        Args:
            name: Name of the Certificate object.
            parent_folder: :ref:`config_object` or :ref:`dn` of the parent folder.
            description: Description of the certificate object.
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` of the contacts for this certificate.
            approvers: List of :ref:`identity_object` or :ref:`prefixed_name` of the approvers for this certificate.
            management_type: Certificate management type.
            service_generated_csr: If ``True``, TPP generates the CSR.
            generate_key_on_application: If ``True``, the key/CSR are generated on the target application.
            hash_algorithm: Hash algorithm.
            common_name: Common name (CN) of the certificate.
            organization: Organization of the certificate.
            organization_unit: Organization units (OU) of the certificate.
            city: City.
            state: State.
            country: Country code.
            san_dns: List of Subject Alternative Names for DNS.
            san_email: List of Subject Alternative Names for e-mail addresses.
            san_upn: List of Subject Alternative Names for UPN.
            san_ip: List of Subject Alternative Names for IP addresses.
            san_uri: List of Subject Alternative Names for URI.
            key_algorithm: Signing key algorithm.
            key_strength: Key strength in bits.
            elliptic_curve: Elliptic curve.
            ca_template: :ref:`config_object` or :ref:`dn` of the Certificate Authority template.
            disable_automatic_renewal: If ``True``, disables automatic renewal.
            renewal_window: Number of days that make the renewal window.
            attributes: Additional attributes that define this certificate.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object`
        """
        cert_attrs = {
            X509CertificateAttributes.description: description,
            X509CertificateAttributes.contact: [self._get_prefixed_universal(c) for c in contacts] if contacts else None,
            X509CertificateAttributes.approver: [self._get_prefixed_universal(a) for a in approvers] if approvers else None,
            X509CertificateAttributes.management_type: management_type,
            X509CertificateAttributes.manual_csr: {True: "0", False: "1"}.get(service_generated_csr),
            X509CertificateAttributes.generate_keypair_on_application: {True: "1", False: "0"}.get(generate_key_on_application),
            X509CertificateAttributes.pkcs10_hash_algorithm: hash_algorithm,
            X509CertificateAttributes.x509_subject: common_name,
            X509CertificateAttributes.organization: organization,
            X509CertificateAttributes.organizational_unit: organization_unit,
            X509CertificateAttributes.city: city,
            X509CertificateAttributes.state: state,
            X509CertificateAttributes.country: country,
            X509CertificateAttributes.driver_name: 'appx509certificate',
            X509CertificateAttributes.x509_subjectaltname_dns: san_dns,
            X509CertificateAttributes.x509_subjectaltname_ipaddress: san_ip,
            X509CertificateAttributes.x509_subjectaltname_uri: san_uri,
            X509CertificateAttributes.x509_subjectaltname_othername_upn: san_upn,
            X509CertificateAttributes.x509_subjectaltname_rfc822: san_email,
            X509CertificateAttributes.key_algorithm: key_algorithm,
            X509CertificateAttributes.key_bit_strength: key_strength,
            X509CertificateAttributes.elliptic_curve: elliptic_curve,
            X509CertificateAttributes.certificate_authority: self._get_dn(ca_template) if ca_template else None,
            X509CertificateAttributes.disable_automatic_renewal: {True: "1", False: "0"}.get(disable_automatic_renewal),
            X509CertificateAttributes.renewal_window: renewal_window
        }
        if attributes:
            cert_attrs.update(attributes)
        return self._config_create(
            name=name,
            parent_folder_dn=self._get_dn(parent_folder),
            config_class=Classes.x509_certificate,
            attributes=cert_attrs,
            get_if_already_exists=get_if_already_exists
        )

    def delete(self, certificate: 'Union[Config.Object, str]'):
        """
        Deletes the certificate object from TPP.

        Args:
            certificate: :ref:`config_object` or :ref:`dn` of the certificate object.
        """
        certificate_guid = self._get_guid(certificate)
        result = self._api.websdk.Certificates.Guid(certificate_guid).delete()
        if not result.is_valid_response():
            certificate_dn = self._get_dn(certificate)
            raise FeatureException(f'Could not delete certificate {certificate_dn}.')

    def details(self, certificate: 'Union[Config.Object, str]'):
        """
        Args:
            certificate: :ref:`config_object` or :ref:`dn` of the certificate object.

        Returns:
            :class:`~.dataclasses.certificate.CertificateDetails`
        """
        return self._get(certificate=certificate).certificate_details

    def dissociate_application(self, certificate: 'Union[Config.Object, str]', applications: 'List[Union[Config.Object, str]]',
                               delete_orphans: bool = False):
        """
        Dissociate an application object from a certificate.

        Args:
            certificate: :ref:`config_object` or :ref:`dn` of the certificate object.
            applications: A list of :ref:`config_object` or :ref:`dn` to each application object.
            delete_orphans: Delete the application object. The corresponding device object will be delete if it has no
                            child applications. Use this option to completely remove the application object and corresponding
                            device objects.
        """
        certificate_dn = self._get_dn(certificate)
        result = self._api.websdk.Certificates.Dissociate.post(
            certificate_dn=certificate_dn,
            application_dn=[self._get_dn(a) for a in applications],
            delete_orphans=delete_orphans
        )
        result.assert_valid_response()

    def download(self, format: str, certificate: 'Union[Config.Object, str]' = None, friendly_name: str = None,
                 include_chain: bool = False, include_private_key: bool = False, keystore_password: str = None,
                 password: str = None, root_first_order: bool = False, vault_id: int = None):
        """
        Downloads a certificate and returns the encoded content, filename, and format as a single object. If ``vault_id``
        is provided, then that specific version of a certificate is downloaded, which is particularly useful when
        trying to download historical certificates.

        Args:
            format: One of the following:
                * Base64
                * Base64 (PKCS #8)
                * DER
                * JKS
                * PKCS #7
                * PKCS #12
            certificate: :ref:`config_object` or :ref:`dn` of the certificate object. Not required if using vault id.
            friendly_name: Label or alias for the given format.
            include_chain: Include parent or root chain.
            include_private_key: Include the private key.
            keystore_password: JKS Keystore password.
            password: Password.
            root_first_order: If ``True``, show root certificate first, followed by intermediate, and finally the
                end entity certificate.
            vault_id: If provided, downloads the certificate with the given Vault ID. Use this when trying
                to download historical certificates. Not required if using certificate config object.

        Returns:
            A *DownloadedCertificate* with these properties

            * **certificate_data** *str* - Encoded certificate data.
            * **format** *str* - File Format.
            * **filename** *str* - File name.
        """
        if vault_id:
            result = self._api.websdk.Certificates.Retrieve.VaultId(vault_id).post(
                format=format,
                friendly_name=friendly_name,
                include_chain=include_chain,
                include_private_key=include_private_key,
                keystore_password=keystore_password,
                password=password,
                root_first_order=root_first_order
            )
        else:
            certificate_dn = self._get_dn(certificate)
            result = self._api.websdk.Certificates.Retrieve.post(
                certificate_dn=certificate_dn,
                format=format,
                friendly_name=friendly_name,
                include_chain=include_chain,
                include_private_key=include_private_key,
                keystore_password=keystore_password,
                password=password,
                root_first_order=root_first_order
            )

        return DownloadedCertificate(
            certificate_data=result.certificate_data,
            filename=result.filename,
            format=result.format
        )

    def get(self, certificate_dn: str, raise_error_if_not_exists: bool = True):
        """
        Args:
            certificate_dn: :ref:`dn` of the certificate object.
            raise_error_if_not_exists: Raise an exception if the object DN does not exist.

        Returns:
            :ref:`config_object`
        """
        return self._get_config_object(object_dn=certificate_dn, raise_error_if_not_exists=raise_error_if_not_exists)

    def get_previous_versions(self, certificate: 'Union[Config.Object, str]', exclude_expired: bool = False,
                              exclude_revoked: bool = False):
        """
        Args:
            certificate: :ref:`config_object` or :ref:`dn` of the certificate object.
            exclude_expired: If ``True``, do not include expired certificates.
            exclude_revoked: If ``True``, do not include revoked certificates.

        Returns:
            List[:class:`~.dataclasses.certificate.PreviousVersions`]
        """
        certificate_guid = self._get_guid(certificate)
        result = self._api.websdk.Certificates.Guid(certificate_guid).PreviousVersions.get(
            exclude_expired=exclude_expired,
            exclude_revoked=exclude_revoked
        )
        return result.previous_versions

    def get_validation_results(self, certificate: 'Union[Config.Object, str]'):
        """
        Args:
            certificate: :ref:`config_object` or :ref:`dn` of the certificate object.

        Returns:
            Tuple[:class:`~.dataclasses.certificate.File`, :class:`~.dataclasses.certificate.SslTls`]
        """
        certificate_guid = self._get_guid(certificate)
        result = self._api.websdk.Certificates.Guid(certificate_guid).ValidationResults.get()
        return result.file, result.ssl_tls

    def push_to_applications(self, certificate: 'Union[Config.Object, str]', applications: 'List[Union[Config.Object, str]]' = None):
        """
        Pushes the active ``certificate`` to the ``applications``.

        Args:
            certificate: :ref:`config_object` or :ref:`dn` of the certificate object.
            applications: A list of :ref:`config_object` or :ref:`dn` to each application object.
        """
        if not applications:
            certificate_dn = self._get_dn(certificate)
            applications = self._api.websdk.Config.ReadDn.post(
                object_dn=certificate_dn,
                attribute_name=X509CertificateAttributes.consumers
            ).values

        self.associate_application(
            certificate=certificate,
            applications=applications,
            push_to_new=True
        )

    def renew(self, certificate: 'Union[Config.Object, str]', csr: str = None, re_enable: bool = False):
        """
        Renews or requests a certificate.

        Args:
            certificate: :ref:`config_object` or :ref:`dn` of the certificate object.
            csr: If provided, uploads the PKCS10 CSR to TPP to send to the CA. If not provided, TPP generates the CSR.
            re_enable: The action to control a previously disabled certificate:
                If ``False``, do not renew a previously disabled certificate.
                If ``True``, clear the Disabled attribute, re-enable, and then renew the certificate (in this request).

        Returns:
            str: The current thumbprint of the active certificate. This should be used when checking the renewal status \
                 to ensure that TPP has registered a new certificate to its vault with a new thumbprint.
        """
        certificate = self._get(certificate)
        current_thumbprint = certificate.certificate_details.thumbprint
        result = self._api.websdk.Certificates.Renew.post(certificate_dn=certificate.dn, pkcs10=csr, reenable=re_enable)
        result.assert_valid_response()
        return current_thumbprint

    def reset(self, certificate: 'Union[Config.Object, str]'):
        """
        Resets the certificate to a non-processing state. No attempt to reprocess the certificate renewal is made.

        Args:
            certificate: :ref:`config_object` or :ref:`dn` of the certificate object.
        """
        certificate_dn = self._get_dn(certificate)
        result = self._api.websdk.Certificates.Reset.post(certificate_dn=certificate_dn, restart=False)
        result.assert_valid_response()
        if not result.processing_reset_completed:
            raise UnexpectedValue(f'Processing reset was not completed for {certificate_dn}.')

    def retry_from_current_stage(self, certificate: 'Union[Config.Object, str]'):
        """
        Retries renewal from the current processing stage of the ``certificate``.

        Args:
            certificate: :ref:`config_object` or :ref:`dn` of the certificate object.
        """
        certificate_dn = self._get_dn(certificate)
        result = self._api.websdk.Certificates.Retry.post(certificate_dn=certificate_dn)
        result.assert_valid_response()

    def retry_from_stage_0(self, certificate: 'Union[Config.Object, str]'):
        """
        Retries renewal from stage 0. This clears all current processing data and restarts
        processing.

        Args:
            certificate: :ref:`config_object` or :ref:`dn` of the certificate object.
        """
        certificate_dn = self._get_dn(certificate)
        result = self._api.websdk.Certificates.Reset.post(certificate_dn=certificate_dn, restart=True)
        result.assert_valid_response()
        if not result.restart_completed:
            raise UnexpectedValue(f'Restart renewal from stage 0 was not triggered on {certificate_dn}.')

    def revoke(self, certificate: 'Union[Config.Object, str]', comments: str = None, disable: bool = None,
               reason: int = None, thumbprint: str = None):
        """
        Revokes the ``certificate``. If a thumbprint is provided, then the particular historical certificate
        associated to the certificate having that thumbprint will be revoked.

        Args:
            certificate: :ref:`config_object` or :ref:`dn` of the certificate object.
            comments: Any comments to include in the revoke request.
            disable: If ``True``, disables the certificate object.
            reason: Reason for revoking.
            thumbprint: If given, the thumbprint of the historical certificate to be revoked.
        """
        certificate_dn = self._get_dn(certificate)
        result = self._api.websdk.Certificates.Revoke.post(
            certificate_dn=certificate_dn,
            comments=comments,
            disable=disable,
            reason=reason,
            thumbprint=thumbprint
        )
        result.assert_valid_response()
        if result.success is not True:
            raise UnexpectedValue(
                f'Cannot revoke {certificate_dn} due to this error:\n{result.error}.'
            )

    def upload(self, certificate_data: str, parent_folder: 'Union[Config.Object, str]', certificate_authority_attributes: dict = None,
               name: str = None, password: str = None, private_key_data: str = None, reconcile: bool = False):
        """
        Uploads the certificate data to TPP to create a certificate object under the given parent folder DN. If the BEGIN/END
        header or footer is missing, the data is assumed to be Base 64 encoded in the PKCS#12 format. For Base 64 encoded
        certificates, characters, such as spaces and new line escape characters (/n), are optional. White space characters are
        removed before any attempt is made to decode the certificate.

        Args:
            certificate_data: Encoded certificate data.
            parent_folder: :ref:`config_object` or :ref:`dn` of the parent folder.
            certificate_authority_attributes: Attributes pertaining to the Certificate Authority to store with the certificate
                object. This is not a DN to a Certificate Authority in TPP.
            name: If given, the name of the new certificate object. If not given, then the Common Name is used.
            password: Password to decrypt the private key.
            private_key_data: Encoded private key data.
            reconcile: If ``False``, replaces the current certificate, if it exists, and stores the current certificate as a
                historical certificate. If ``True``, then TPP activates the certificate with the newest "ValidFrom" date and
                archives the other certificate as a historical certificate.

        Returns:
            :ref:`config_object` of the uploaded certificate.
        """
        if certificate_authority_attributes:
            certificate_authority_attributes = self._name_value_list(attributes=certificate_authority_attributes)

        result = self._api.websdk.Certificates.Import.post(
            certificate_data=certificate_data,
            policy_dn=self._get_dn(parent_folder),
            ca_specific_attributes=certificate_authority_attributes,
            object_name=name,
            password=password,
            private_key_data=private_key_data,
            reconcile=reconcile
        )
        result.assert_valid_response()
        return self._api.websdk.Config.IsValid.post(object_dn=result.certificate_dn).object

    def validate(self, certificates: 'List[Union[Config.Object, str]]'):
        """
        Performs SSL/TLS network validation of certificate on all applications associated to certificate that are not disabled.

        Args:
            certificates: List of :ref:`config_object` or :ref:`dn` to validate.

        Returns:
            Tuple[str, List[str]]: Tuple of :ref:`dn` and validation warnings
        """
        certificate_dns = [self._get_dn(certificate) for certificate in certificates]
        result = self._api.websdk.Certificates.Validate.post(certificate_dns=certificate_dns)
        if not result.success:
            raise UnexpectedValue(
                f'TPP failed to perform validation. Got this error: "{result.error}".'
            )
        return result.validated_certificate_dns, result.warnings

    def wait_for_enrollment_to_complete(self, certificate: 'Union[Config.Object, str]', current_thumbprint: str,
                                        timeout: int = 60, poll_interval: int = 1):
        """
        Waits for the certificate renewal to complete over a period of ``timeout`` seconds. The ``current_thumbprint``
        is returned by :meth:`renew`. Renewal is complete when the ``current_thumbprint`` does not match the new
        thumbprint and either the processing stage is "None" or greater than or equal to 800, which is the start of the
        provisioning stage. If the certificate management type is set to *Provisioning*, use the application feature
        ``wait_for_installation_to_complete()``.

        Args:
            certificate: :ref:`config_object` or :ref:`dn` of the certificate object.
            current_thumbprint: Thumbprint of the `current` certificate object.
            timeout: Timeout in seconds before raising an error.
            poll_interval: Interval to poll TPP for the renewal status.

        Returns:
            :class:`~.dataclasses.certificate.CertificateDetails`
        """
        cert = self._get(certificate=certificate)
        with self._Timeout(timeout=timeout) as to:
            while not to.is_expired(poll=poll_interval):
                cert.assert_valid_response()
                thumbprint = cert.certificate_details.thumbprint
                if thumbprint and thumbprint != current_thumbprint and cert.processing_details.stage in {None, 800}:
                    return cert.certificate_details
                elif cert.processing_details.in_error:
                    break
                cert = self._get(certificate=certificate)

        raise TimeoutError(
            f'Certificate renewal for "{cert.dn}" encountered an error at stage {cert.processing_details.stage} with '
            f'status "{cert.processing_details.status}".'
        )

    def wait_for_stage(self, certificate: 'Union[Config.Object, str]', stage: int, expect_workflow: bool = True,
                       timeout: int = 60, poll_interval: int = 1):
        """
        Waits for the current processing of the certificate to reach the given ``stage`` over a period of
        ``timeout`` seconds. If the timeout is reached, an error is raised.

        Args:
            certificate: :ref:`config_object` or :ref:`dn` of the certificate object.
            stage: Stage at which to return
            expect_workflow: If ``True``, validates that a Ticket DN has been issued to the certificate.
            timeout: Timeout in seconds before throwing an error.
            poll_interval: Interval to poll TPP for the renewal status.

        Returns:
            The values returned by the |Websdk|, namely

            * **approver** *(List[str])* - List of approvers on the certificate object.
            * **certificate_details** (:class:`~.dataclasses.certificate.CertificateDetails`) - Certificate details.
            * **contact** *(List[str])* - List of contacts on the certificate object.
            * **created_on** *(datetime)* - Date on which the certificate object was created.
            * **custom_fields** *(List[dict])* - Custom fields on the certificate object.
            * **dn** *(str)* - :ref:`dn` of the certificate object.
            * **guid** *(str)* - :ref:`guid` of the certificate object.
            * **name** *(str)* - Name of the certificate object.
            * **parent_dn** *(str)* - Parent :ref:`dn` of the certificate object.
            * **processing_details** (:class:`~.dataclasses.certificate.ProcessingDetails`) - Certificate processing details.
            * **renewal_details** (:class:`~.dataclasses.certificate.RenewalDetails`) - Certificate renewal settings details.
            * **schema_class** *(str)* - Schema class.
            * **validation_details** (:class:`~.dataclasses.certificate.ValidationDetails`) - Certificate validation details.
        """
        cert = self._get(certificate=certificate)
        with self._Timeout(timeout=timeout) as to:
            while not to.is_expired(poll=poll_interval):
                if cert.processing_details.in_error is True:
                    break
                if cert.processing_details.stage is not None:
                    if expect_workflow and cert.processing_details.stage == stage:
                        if 'pending workflow resolution' in cert.processing_details.status.lower():
                            return cert
                    elif not expect_workflow and cert.processing_details.stage >= stage:
                        return cert
                cert = self._get(certificate=certificate)

        if expect_workflow and stage == cert.processing_details.stage:
            msg = f'After {timeout} seconds certificate renewal reached stage {stage}, but no workflow was issued. ' \
                  f'The current stage is {cert.processing_details.stage} with status {cert.processing_details.status}.'
        else:
            msg = f'Certificate renewal did not reach stage {stage} after {timeout} seconds. The current ' \
                  f'stage is {cert.processing_details.stage} with status {cert.processing_details.status}.'
        raise UnexpectedValue(msg)



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.top import TopAttributes


class ACMEOrderAttributes(TopAttributes, metaclass=IterableMeta):
	__config_class__ = "ACME Order"
	acme_authorization_dn = Attribute('ACME Authorization DN', min_version='17.2')
	acme_expires = Attribute('ACME Expires', min_version='17.2')
	acme_not_after = Attribute('ACME Not After', min_version='17.2')
	acme_not_before = Attribute('ACME Not Before', min_version='17.2')
	csr_vault_id = Attribute('CSR Vault Id', min_version='17.2')
	certificate = Attribute('Certificate', min_version='17.2')
	error = Attribute('Error', min_version='17.2')
	identifier = Attribute('Identifier', min_version='21.2')
	status = Attribute('Status', min_version='17.2')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.log_channel import LogChannelAttributes


class LogSplunkAttributes(LogChannelAttributes, metaclass=IterableMeta):
	__config_class__ = "Log Splunk"
	credential = Attribute('Credential')
	host = Attribute('Host')
	index = Attribute('Index')
	port = Attribute('Port')
	source = Attribute('Source')
	timeout = Attribute('Timeout')
	verbose = Attribute('Verbose')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.application_base import ApplicationBaseAttributes


class BasicAttributes(ApplicationBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Basic"
	certificate_file = Attribute('Certificate File', min_version='15.2')
	network_validation_disabled = Attribute('Network Validation Disabled')



from pytpp.attributes._helper import IterableMeta
from pytpp.attributes.credential_driver_base import CredentialDriverBaseAttributes


class IntegratedCredentialDriverAttributes(CredentialDriverBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Integrated Credential Driver"



from typing import List 
from pytpp.api.api_base import API, APIResponse, api_response_property
from pytpp.properties.response_objects.secret_store import SecretStore


class _SecretStore:
    def __init__(self, api_obj):
        self.Add = self._Add(api_obj=api_obj)
        self.Associate = self._Associate(api_obj=api_obj)
        self.Dissociate = self._Dissociate(api_obj=api_obj)
        self.EncryptionKeysInUse = self._EncryptionKeysInUse(api_obj=api_obj)
        self.Lookup = self._Lookup(api_obj=api_obj)
        self.LookupAllAssociationsbyVaultid = self._LookupAllAssociationsbyVaultid(api_obj=api_obj)
        self.LookupByAssociation = self._LookupByAssociation(api_obj=api_obj)
        self.LookupAssociationbyVaultID = self._LookupAssociationbyVaultID(api_obj=api_obj)
        self.LookupByOwner = self._LookupByOwner(api_obj=api_obj)
        self.LookupByVaultType = self._LookupByVaultType(api_obj=api_obj)
        self.Mutate = self._Mutate(api_obj=api_obj)
        self.OrphanLookup = self._OrphanLookup(api_obj=api_obj)
        self.OwnerAdd = self._OwnerAdd(api_obj=api_obj)
        self.OwnerDelete = self._OwnerDelete(api_obj=api_obj)
        self.OwnerLookup = self._OwnerLookup(api_obj=api_obj)
        self.Retrieve = self._Retrieve(api_obj=api_obj)

    class _Add(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SecretStore/Add')

        def post(self, base_64_data: str, keyname: str, namespace: str, owner: str, vault_type: int):
            body = {
                'Base64Data': base_64_data,
                'Keyname': keyname,
                'Namespace': namespace,
                'Owner': owner,
                'VaultType': vault_type
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def result(self):
                    return SecretStore.Result(self._from_json(key='Result'))

                @property
                @api_response_property()
                def vault_id(self) -> int:
                    return self._from_json(key='VaultID')

            return _Response(response=self._post(data=body))

    class _Associate(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SecretStore/Associate')

        def post(self, name: str, vault_id: int, date_value: int = None, int_value: int = None, string_value: str = None):
            body = {
                'Name': name,
                'VaultID': vault_id,
                'DateValue': date_value,
                'IntValue': int_value,
                'StringValue': string_value
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def result(self):
                    return SecretStore.Result(self._from_json(key='Result'))

            return _Response(response=self._post(data=body))

    class _Dissociate(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SecretStore/Dissociate')

        def post(self, vault_id: int, int_value: int = None, name: str = None, string_value: str = None, date_value: int = None):
            body = {
                'VaultID': vault_id,
                'IntValue': int_value,
                'Name': name,
                'StringValue': string_value,
                'DateValue': date_value
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def result(self):
                    return SecretStore.Result(self._from_json(key='Result'))

            return _Response(response=self._post(data=body))

    class _EncryptionKeysInUse(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SecretStore/EncryptionKeysInUse')

        def get(self):
            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def encryption_keys(self) -> List[str]:
                    return self._from_json(key='EncryptionKeys')

                @property
                @api_response_property()
                def result(self):
                    return SecretStore.Result(self._from_json(key='Result'))

            return _Response(response=self._get())

    class _Lookup(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SecretStore/Lookup')

        def get(self):
            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def result(self):
                    return SecretStore.Result(self._from_json(key='Result'))

                @property
                @api_response_property()
                def vault_ids(self) -> List[int]:
                    return self._from_json(key='VaultIDs')

            return _Response(response=self._get())

    class _LookupAllAssociationsbyVaultid(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SecretStore/LookupAllAssociationsbyVaultid')

        def post(self, vault_id: int):
            body = {
                'VaultID': vault_id
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def result(self):
                    return SecretStore.Result(self._from_json(key='Result'))

                @property
                @api_response_property()
                def typed_name_values(self):
                    return [SecretStore.TypedNameValues(tnv) for tnv in self._from_json('TypedNameValues')]

            return _Response(response=self._post(data=body))

    class _LookupByAssociation(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SecretStore/LookupByAssociation')

        def post(self, name: str, int_value: int = None ,string_value: str = None, date_value: int = None):
            body = {
                'Name': name,
                'IntValue': int_value,
                'StringValue': string_value,
                'DateValue': date_value
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def result(self):
                    return SecretStore.Result(self._from_json(key='Result'))

                @property
                @api_response_property()
                def vault_ids(self) -> List[int]:
                    return self._from_json(key='VaultIDs')

            return _Response(response=self._post(data=body))

    class _LookupAssociationbyVaultID(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SecretStore/LookupAssociationbyVaultID')

        def post(self, vault_id: int, name: str = None):
            body = {
                'VaultID': vault_id,
                'Name': name
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def result(self):
                    return SecretStore.Result(self._from_json(key='Result'))

                @property
                @api_response_property()
                def value(self) -> str:
                    return self._from_json(key='Value')

            return _Response(response=self._post(data=body))

    class _LookupByOwner(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SecretStore/LookupByOwner')

        def post(self, namespace: str, owner: str, vault_type: str = None):
            body = {
                'Namespace': namespace,
                'Owner': owner,
                'VaultType': vault_type
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def result(self):
                    return SecretStore.Result(self._from_json(key='Result'))

                @property
                @api_response_property()
                def vault_ids(self) -> List[int]:
                    return self._from_json(key='VaultIDs')

            return _Response(response=self._post(data=body))

    class _LookupByVaultType(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SecretStore/LookupByVaultType')

        def post(self, vault_type: int):
            body = {
                'VaultType': vault_type
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def result(self):
                    return SecretStore.Result(self._from_json(key='Result'))

                @property
                @api_response_property()
                def vault_ids(self) -> List[int]:
                    return self._from_json(key='VaultIDs')

            return _Response(response=self._post(data=body))

    class _Mutate(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SecretStore/Mutate')

        def post(self, vault_id: int, vault_type: int):
            body = {
                'VaultID': vault_id,
                'VaultType': vault_type
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def result(self):
                    return SecretStore.Result(self._from_json(key='Result'))

            return _Response(response=self._post(data=body))

    class _OrphanLookup(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SecretStore/OrphanLookup')

        def post(self, vault_type: int):
            body = {
                'VaultType': vault_type
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def result(self):
                    return SecretStore.Result(self._from_json(key='Result'))

                @property
                @api_response_property()
                def vault_ids(self) -> List[int]:
                    return self._from_json(key='VaultIDs')

            return _Response(response=self._post(data=body))

    class _OwnerAdd(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SecretStore/OwnerAdd')

        def post(self, namespace: str, owner: str, vault_id: int):
            body = {
                'Namespace': namespace,
                'Owner': owner,
                'VaultId': vault_id
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def result(self):
                    return SecretStore.Result(self._from_json(key='Result'))

            return _Response(response=self._post(data=body))

    class _OwnerDelete(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SecretStore/OwnerDelete')

        def post(self, namespace: str, owner: str, vault_id: int = None):
            body = {
                'Namespace': namespace,
                'Owner': owner,
                'VaultId': vault_id
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def result(self):
                    return SecretStore.Result(self._from_json(key='Result'))

            return _Response(response=self._post(data=body))

    class _OwnerLookup(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SecretStore/OwnerLookup')

        def post(self, namespace: str, vault_id: int):
            body = {
                'Namespace': namespace,
                'VaultID': vault_id
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def result(self):
                    return SecretStore.Result(self._from_json(key='Result'))

                @property
                @api_response_property()
                def owners(self) -> List[str]:
                    return self._from_json(key='Owners')

            return _Response(response=self._post(data=body))

    class _Retrieve(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SecretStore/Retrieve')

        def post(self, vault_id: int):
            body = {
                'VaultID': vault_id
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def base_64_data(self) -> str:
                    return self._from_json(key='Base64Data')

                @property
                @api_response_property()
                def result(self):
                    return SecretStore.Result(self._from_json(key='Result'))

                @property
                @api_response_property()
                def vault_type(self) -> str:
                    return self._from_json(key='VaultType')

            return _Response(response=self._post(data=body))



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.schedule_base import ScheduleBaseAttributes
from pytpp.attributes.top import TopAttributes


class PlacementJobAttributes(ScheduleBaseAttributes, TopAttributes, metaclass=IterableMeta):
	__config_class__ = "Placement Job"
	archive_old_certificates = Attribute('Archive Old Certificates', min_version='19.1')
	default_container = Attribute('Default Container', min_version='19.1')
	include_subfolders = Attribute('Include Subfolders', min_version='19.1')
	last_run = Attribute('Last Run', min_version='19.1')
	layout_rules = Attribute('Layout Rules', min_version='19.1')
	objects_combined = Attribute('Objects Combined', min_version='19.1')
	objects_evaluated = Attribute('Objects Evaluated', min_version='19.1')
	objects_moved = Attribute('Objects Moved', min_version='19.1')
	progress = Attribute('Progress', min_version='19.1')
	rules_order = Attribute('Rules Order', min_version='19.1')
	scan_folders = Attribute('Scan Folders', min_version='19.1')
	status = Attribute('Status', min_version='19.1')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.service_module import ServiceModuleAttributes


class KmipServiceModuleAttributes(ServiceModuleAttributes, metaclass=IterableMeta):
	__config_class__ = "Kmip Service Module"
	agent_module_handlers = Attribute('Agent Module Handlers')
	agent_upgrade_rule = Attribute('Agent Upgrade Rule')
	credential = Attribute('Credential')
	interval = Attribute('Interval')
	port = Attribute('Port')
	start_time = Attribute('Start Time')
	upgrade = Attribute('Upgrade')
	version = Attribute('Version')



from dataclasses import dataclass
from typing import List, Union
from pytpp.features.bases.feature_base import FeatureBase, feature
from pytpp.features.definitions.exceptions import InvalidResultCode, InvalidFormat
from pytpp.tools.vtypes import Config


@dataclass
class AttributeValue:
    values: List[str]
    locked: bool


@feature('Objects')
class Objects(FeatureBase):
    def __init__(self, api):
        super().__init__(api=api)

    def clear(self, obj: 'Union[Config.Object, str]', attributes: Union[dict, List[str]]):
        """
        Clears attributes from an object.
        
        Args:
            obj: :ref:`config_object` or :ref:`dn` of the object.
            attributes: Two types are supported:

                * ``list`` of attribute names to be cleared entirely.
                * ``dict`` whose keys are attribute names and whose values are the values to be cleared
                  from the attribute. If the attribute is left empty it is cleared.
        """
        obj_dn = self._get_dn(obj)
        if isinstance(attributes, list):
            for attribute_name in attributes:
                result = self._api.websdk.Config.ClearAttribute.post(
                    object_dn=obj_dn,
                    attribute_name=attribute_name
                ).result

                if result.code != 1:
                    raise InvalidResultCode(code=result.code, code_description=result.config_result)

        elif isinstance(attributes, dict):
            for name, values in attributes.items():
                if not isinstance(values, list):
                    values = [values]

                for value in values:
                    result = self._api.websdk.Config.RemoveDnValue.post(
                        object_dn=obj_dn,
                        attribute_name=name,
                        value=value
                    ).result

                    if result.code != 1:
                        raise InvalidResultCode(code=result.code, code_description=result.config_result)
        else:
            raise TypeError(f'Expected attributes to be of type "list" or "dict", but got {type(attributes)} instead.')

    def exists(self, object_dn: str = None):
        """
        Args:
            object_dn: The :ref:`dn` of the object.

        Returns:
           bool: ``True`` if ``object_dn`` exist, else ``False``.
        """
        dn = self._get_dn(obj=object_dn)
        result = self._api.websdk.Config.IsValid.post(object_dn=dn)
        return result.is_valid_response() and result.result.code == 1

    def find_policy(self, obj: 'Union[Config.Object, str]', class_name: str, attribute_name: str):
        """
        Find the folder that defines the policy attribute of the ``class_name`` on a given object.

        Args:
            obj: :ref:`config_object` or :ref:`dn` of the object.
            class_name: TPP class name of the object.
            attribute_name: Name of the attribute.

        Returns:
            Tuple[str, List[str], bool]: A ``tuple`` of

                * A :ref:`dn` of the folder implementing the policy.
                * A list of attribute values.
                * ``True`` if the policy value is locked, else ``False``.
        """
        obj_dn = self._get_dn(obj)
        resp = self._api.websdk.Config.FindPolicy.post(
            object_dn=obj_dn,
            class_name=str(class_name),
            attribute_name=attribute_name
        )

        result = resp.result
        if result.code != 1:
            raise InvalidResultCode(code=result.code, code_description=result.config_result)

        class Policy:
            policy_dn = resp.policy_dn
            values = resp.values
            locked = resp.locked

        return Policy()

    def get(self, object_dn: str = None, object_guid: str = None, raise_error_if_not_exists: bool = True):
        """
        One of ``object_dn`` or ``object_guid`` is required.

        Args:
            object_dn: :ref:`dn` of the object.
            object_guid: GUID of the object.
            raise_error_if_not_exists: If ``True`` raise an exception if the obejct doesn't exist.

        Returns:
            :ref:`config_object` of the object.
        """
        return self._get_config_object(
            object_dn=object_dn,
            object_guid=object_guid,
            raise_error_if_not_exists=raise_error_if_not_exists
        )

    def read(self, obj: 'Union[Config.Object, str]', attribute_name: str, include_policy_values: bool = False, timeout: int = 10):
        """
        Args:
            obj: :ref:`config_object` or :ref:`dn` of the object.
            attribute_name: The attribute name.
            include_policy_values: If ``True``, the effective value(s) are returned.
                Otherwise only values explicitly assigned to the object are returned.
            timeout: Read timeout in seconds.

        Returns:
            An ``AttributeValue`` object with these properties

            * **values** *(List[str])* - List of attribute values.
            * **locked** *(bool)* - ``True`` if the value is locked by policy.
        """
        with self._Timeout(timeout=timeout) as to:
            while not to.is_expired():
                result, attr = self._read(
                    obj=obj,
                    attribute_name=attribute_name,
                    include_policy_values=include_policy_values
                )
                if result.code != 1:
                    continue
                if attr and attr.values:
                    return attr

        obj_dn = self._get_dn(obj)
        InvalidResultCode(code=result.code, code_description=result.config_result).log()
        raise TimeoutError(f'Could not read {attribute_name} on {obj_dn} because it did not exist '
                           f'after {timeout} seconds.')

    def read_all(self, obj: 'Union[Config.Object, str]'):
        """
        Args:
            obj: :ref:`config_object` or :ref:`dn` of the object.

        Returns:
            List of :class:`~.dataclasses.config.NameValues` where the

            * name is the attribute name.
            * values is the list of attribute values.
        """
        obj_dn = self._get_dn(obj)
        resp = self._api.websdk.Config.ReadAll.post(object_dn=obj_dn)

        result = resp.result
        if result.code != 1:
            InvalidResultCode(code=result.code, code_description=result.config_result).log()

        return resp.name_values

    def rename(self, obj: 'Union[Config.Object, str]', new_object_dn: str):
        """
        .. note::
            This method can be used to rename objects and move their location.

        Args:
            obj: :ref:`config_object` or :ref:`dn` of the object.
            new_object_dn: :ref:`dn` of the new object.
        """
        obj_dn = self._get_dn(obj)
        if not new_object_dn.startswith('\\VED'):
            raise InvalidFormat(f'"{new_object_dn}" must be an absolute path starting from \\VED.')

        response = self._api.websdk.Config.RenameObject.post(object_dn=obj_dn, new_object_dn=new_object_dn)
        result = response.result

        if result.code != 1:
            raise InvalidResultCode(code=result.code, code_description=result.config_result)

        return self.get(object_dn=new_object_dn, raise_error_if_not_exists=True)

    def update(self, obj: 'Union[Config.Object, str]', attributes: dict):
        """
        Updates attributes on an object. If the attribute is locked TPP will simply ignore the request.

        Args:
            obj: :ref:`config_object` or :ref:`dn` of the object.
            attributes: A dictionary of attribute name/value pairs where the name is the
                attribute name and the value is the attribute value.
        """
        obj_dn = self._get_dn(obj)
        for name, value in attributes.items():
            result = self._api.websdk.Config.AddValue.post(
                object_dn=obj_dn,
                attribute_name=name,
                value=value,
            ).result

            if result.code != 1:
                raise InvalidResultCode(code=result.code, code_description=result.config_result)

    def wait_for(self, obj: 'Union[Config.Object, str]', attribute_name: str, attribute_value: str, include_policy_values: bool = False,
                 timeout: int = 10):
        """
        Waits for the ``attribute_name`` to have the ``attribute_value`` on the object within the timeout period. A
        ``TimeoutError`` is raised if the ``attribute_name`` does not have the ``attribute_value``.

        Args:
            obj: :ref:`config_object` or :ref:`dn` of the object.
            attribute_name: The name of the attribute.
            attribute_value: The expected value to the ``attribute_name``.
            include_policy_values: If ``True``, the effective value(s) are returned.
                Otherwise only values explicitly assigned to the object  are returned.
            timeout: Timeout period in seconds.

        Returns:
            An ``AttributeValue`` object with these properties

            * **values** *(List[str])* - List of attribute values.
            * **locked** *(bool)* - ``True`` if the value is locked by policy.
        """
        obj_dn = self._get_dn(obj)
        with self._Timeout(timeout=timeout) as to:
            while not to.is_expired():
                result, attr = self._read(
                    obj=obj_dn,
                    attribute_name=attribute_name,
                    include_policy_values=include_policy_values
                )
                if result.code != 1:
                    continue
                if attr and any([True for value in attr.values if str(value).lower() == attribute_value.lower()]):
                    return attr

        InvalidResultCode(code=result.code, code_description=result.config_result).log()
        raise TimeoutError(
            f'{attribute_name} on "{obj_dn}" did not return {attribute_value} in {timeout} seconds. '
            f'Got {attr.values} instead.'
        )

    def write(self, obj: 'Union[Config.Object, str]', attributes: dict):
        """
        Writes new attributes on an object. If the attribute is locked TPP will simply ignore the request.

        Args:
            obj: :ref:`config_object` or :ref:`dn` of the object.
            attributes: A dictionary of attribute name/value pairs where the name is the
                attribute name and the value is the attribute value.
        """
        obj_dn = self._get_dn(obj)
        attributes = {k: ([v] if not isinstance(v, list) else v) for k, v in attributes.items()}

        result = self._api.websdk.Config.Write.post(
            object_dn=obj_dn,
            attribute_data=self._name_value_list(attributes, keep_list_values=True)
        ).result

        if result.code != 1:
            raise InvalidResultCode(code=result.code, code_description=result.config_result)

    def _read(self, obj: 'Union[Config.Object, str]', attribute_name: str, include_policy_values: bool):
        obj_dn = self._get_dn(obj)
        if include_policy_values is True:
            resp = self._api.websdk.Config.ReadEffectivePolicy.post(
                object_dn=obj_dn,
                attribute_name=attribute_name
            )
            return resp.result, AttributeValue(values=resp.values, locked=resp.locked)
        else:
            resp = self._api.websdk.Config.Read.post(
                object_dn=obj_dn,
                attribute_name=attribute_name
            )
            return resp.result, AttributeValue(values=resp.values, locked=False)



from pytpp.features.bases.feature_base import FeatureBase, feature
from pytpp.features.definitions.exceptions import InvalidResultCode
from pytpp.features.definitions.classes import Classes
from pytpp.attributes.device import DeviceAttributes
from pytpp.attributes.jump_server import JumpServerAttributes
from typing import Union, List, TYPE_CHECKING
if TYPE_CHECKING:
    from pytpp.tools.vtypes import Config, Identity


class _DeviceBase(FeatureBase):
    def __init__(self, api):
        super().__init__(api=api)

    def delete(self, device: 'Union[Config.Object, str]'):
        """
        Deletes the device object specified.

        Args:
            device: :ref:`config_object` or :ref:`dn` of to the device object.
        """
        dn = self._get_dn(device)
        self._config_delete(object_dn=dn)


@feature('Device')
class Device(_DeviceBase):
    def __init__(self, api):
        super().__init__(api=api)

    def create(self, name: str, parent_folder: 'Union[Config.Object, str]', description: 'str' = None,
               contacts: 'List[Union[Identity.Identity, str]]' = None, address: 'str' = None, agent_provisioning_mode: 'bool' = None,
               concurrent_connection_limit: 'int' = None, device_credential: 'Union[Config.Object, str]' = None,
               temp_directory: 'str' = None, os_type: 'str' = None, jump_server: 'Union[Config.Object, str]' = None,
               use_sudo: 'bool' = None, sudo_credential: 'Union[Config.Object, str]' = None, enforce_host_key: 'bool' = None,
               attributes: dict = None, get_if_already_exists: bool = True):
        """
        Args:
            name: Name of the device object .
            parent_folder: :ref:`config_object` or :ref:`dn` of the parent folder.
            description: Description of the device object.
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` of the contacts.
            address: Hostname or IP address of the device.
            agent_provisioning_mode: Set provisioning mode to connect via Agent if ``True``.
            concurrent_connection_limit: Concurrent connection limit to this device.
            device_credential: :ref:`config_object` or :ref:`dn` of the device credential.
            temp_directory: Temp directory.
            os_type: Operating Sytem type.
            jump_server: :ref:`config_object` or :ref:`dn` of the jump server.
            use_sudo: Use sudo.
            sudo_credential: :ref:`config_object` or :ref:`dn` of the sudo user credential.
            enforce_host_key: Enforce host key policy.
            attributes: List of attributes pertaining to the device object.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the device object.
        """
        dev_attrs = {
            DeviceAttributes.description: description,
            DeviceAttributes.contact: [self._get_prefixed_universal(c) for c in contacts] if contacts else None,
            DeviceAttributes.host: address,
            DeviceAttributes.connection_method: {True: "Agent"}.get(agent_provisioning_mode),
            DeviceAttributes.concurrent_connection_limit: concurrent_connection_limit,
            DeviceAttributes.credential: self._get_dn(device_credential) if device_credential else None,
            DeviceAttributes.temp_directory: temp_directory,
            DeviceAttributes.remote_server_type: os_type,
            DeviceAttributes.jump_server_dn: self._get_dn(jump_server) if jump_server else None,
            DeviceAttributes.global_sudo: {True: "1", False: "0"}.get(use_sudo),
            DeviceAttributes.secondary_credential: self._get_dn(sudo_credential) if sudo_credential else None,
            DeviceAttributes.enforce_known_host: {True: "1", False: "0"}.get(enforce_host_key)
        }
        if attributes:
            dev_attrs.update(attributes)

        return self._config_create(
            name=name,
            parent_folder_dn=self._get_dn(parent_folder),
            config_class=Classes.device,
            attributes=dev_attrs,
            get_if_already_exists=get_if_already_exists
        )

    def get(self, device_dn: str, raise_error_if_not_exists: bool = True):
        """
        Args:
            device_dn: :ref:`dn` of the device object.
            raise_error_if_not_exists: Raise an exception if the device DN does not exist.

        Returns:
            :ref:`config_object` of the device object.
        """
        return self._get_config_object(
            object_dn=device_dn,
            raise_error_if_not_exists=raise_error_if_not_exists
        )

    def scan_for_ssh_keys(self, device: 'Union[Config.Object, str]'):
        """
        Submits Agentless discovery work for the given device.

        Args:
            device: :ref:`config_object` or :ref:`dn` of the device object.
        """
        dn = self._get_dn(device)
        result = self._api.websdk.Config.Write.post(
            object_dn=dn,
            attribute_data=[{"Name": "Agentless Discovery To Do",
                            "Value": "1"}]
        ).result

        if result.code != 1:
            raise InvalidResultCode(code=result.code, code_description=result.config_result)


@feature('Jump Server')
class JumpServer(_DeviceBase):
    def __init__(self, api):
        super().__init__(api=api)

    def create(self, name: str, parent_folder: 'Union[Config.Object, str]', description: 'str' = None,
               contacts: 'List[Union[Identity.Identity, str]]' = None, address: 'str' = None, port: 'int' = None,
               concurrent_connection_limit: 'int' = None, device_credential: 'Union[Config.Object, str]' = None,
               temp_directory: 'str' = None, os_type: 'str' = None, ssh_version: 'str' = None, ssh_syntax: 'str' = None,
               use_sudo: 'bool' = None, sudo_credential: 'Union[Config.Object, str]' = None,
               enforce_host_key: 'bool' = None, attributes: dict = None, get_if_already_exists: bool = True):
        """
        Args:
            name: Name of the device object .
            parent_folder: :ref:`config_object` or :ref:`dn` of the parent folder.
            description: Description of the device object.
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` of the contacts.
            address: Hostname or IP address of the device.
            port: Connection port.
            concurrent_connection_limit: Concurrent connection limit to this device.
            device_credential: :ref:`config_object` or :ref:`dn` of the device credential.
            temp_directory: Temp directory.
            os_type: Operating Sytem type.
            ssh_version: SSH version.
            ssh_syntax: SSH syntax.
            use_sudo: Use sudo.
            sudo_credential: :ref:`config_object` or :ref:`dn` of the sudo user credential.
            enforce_host_key: Enforce host key policy.
            attributes: List of attributes pertaining to the device object.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the jump server object.
        """
        dev_attrs = {
            JumpServerAttributes.description                : description,
            JumpServerAttributes.contact                    : [self._get_prefixed_universal(c) for c in contacts] if contacts else None,
            JumpServerAttributes.host                       : address,
            JumpServerAttributes.port                       : port,
            JumpServerAttributes.concurrent_connection_limit: concurrent_connection_limit,
            JumpServerAttributes.credential                 : self._get_dn(device_credential) if device_credential else None,
            JumpServerAttributes.temp_directory             : temp_directory,
            JumpServerAttributes.remote_server_type         : os_type,
            JumpServerAttributes.ssh_version                : ssh_version,
            JumpServerAttributes.ssh_connection_string      : ssh_syntax,
            JumpServerAttributes.global_sudo                : {True: "1", False: "0"}.get(use_sudo),
            JumpServerAttributes.secondary_credential       : self._get_dn(sudo_credential) if sudo_credential else None,
            JumpServerAttributes.enforce_known_host         : {True: "1", False: "0"}.get(enforce_host_key)
        }
        if attributes:
            dev_attrs.update(attributes)

        return self._config_create(
            name=name,
            parent_folder_dn=self._get_dn(parent_folder),
            config_class=Classes.jump_server,
            attributes=dev_attrs,
            get_if_already_exists=get_if_already_exists
        )

    def get(self, device_dn: str, raise_error_if_not_exists: bool = True):
        """
        Args:
            device_dn: :ref:`dn` of the device object.
            raise_error_if_not_exists: Raise an exception if the device DN does not exist.

        Returns:
            :ref:`config_object` of the device object.
        """
        return self._get_config_object(
            object_dn=device_dn,
            raise_error_if_not_exists=raise_error_if_not_exists
        )

    def scan_for_ssh_keys(self, device: 'Union[Config.Object, str]'):
        """
        Submits Agentless discovery work for the given device.

        Args:
            device: :ref:`config_object` or :ref:`dn` of the device object.
        """
        dn = self._get_dn(device)
        result = self._api.websdk.Config.Write.post(
            object_dn=dn,
            attribute_data=[{
                                "Name" : "Agentless Discovery To Do",
                                "Value": "1"
                            }]
        ).result

        if result.code != 1:
            raise InvalidResultCode(code=result.code, code_description=result.config_result)



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.credential_driver_base import CredentialDriverBaseAttributes


class CyberArkCredentialsDriverAttributes(CredentialDriverBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "CyberArk Credentials Driver"
	central_credential_provider_web_service_url = Attribute('Central Credential Provider Web Service URL', min_version='21.4')
	password_retrieval_method = Attribute('Password Retrieval Method', min_version='21.4')
	scim_server_url = Attribute('SCIM Server URL', min_version='19.3')
	scim_server_user = Attribute('SCIM Server User', min_version='19.3')
	use_proxy = Attribute('Use Proxy', min_version='18.1')
	web_service_url = Attribute('Web Service URL', min_version='17.2')
	web_service_user = Attribute('Web Service User', min_version='17.2')
	windows_credential_provider_version = Attribute('Windows Credential Provider Version', min_version='21.4')



from pytpp.attributes._helper import IterableMeta
from pytpp.attributes.organization import OrganizationAttributes


class PlacementJobContainerAttributes(OrganizationAttributes, metaclass=IterableMeta):
	__config_class__ = "Placement Job Container"



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.application_base import ApplicationBaseAttributes


class RiverbedSteelHeadAttributes(ApplicationBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Riverbed SteelHead"
	certificate_type = Attribute('Certificate Type', min_version='15.2')
	file_validation_disabled = Attribute('File Validation Disabled', min_version='15.2')
	install_chain = Attribute('Install Chain', min_version='15.2')
	network_validation_disabled = Attribute('Network Validation Disabled', min_version='15.2')
	replace_existing = Attribute('Replace Existing', min_version='15.2')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.certificate_trust_store_base import CertificateTrustStoreBaseAttributes


class ConnectDirectTrustStoreAttributes(CertificateTrustStoreBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "ConnectDirect Trust Store"
	key_store = Attribute('Key Store')
	key_store_credential = Attribute('Key Store Credential')
	node_name = Attribute('Node Name')
	protocol = Attribute('Protocol')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.certificate_trust_store_base import CertificateTrustStoreBaseAttributes


class CAPITrustStoreAttributes(CertificateTrustStoreBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "CAPI Trust Store"
	log_windows_events = Attribute('Log Windows Events')
	trust_store_name = Attribute('Trust Store Name', min_version='20.1')



from pytpp.attributes._helper import IterableMeta
from pytpp.attributes.schedule_base import ScheduleBaseAttributes
from pytpp.attributes.top import TopAttributes


class CodeSigningKeyTimeConstraintAttributes(ScheduleBaseAttributes, TopAttributes, metaclass=IterableMeta):
	__config_class__ = "Code Signing Key Time Constraint"



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.branch_base import BranchBaseAttributes


class SecretStoreRootAttributes(BranchBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Secret Store Root"
	certificate_protection_key = Attribute('Certificate Protection Key')
	dsn = Attribute('DSN')
	driver_name = Attribute('Driver Name')
	non_authoritative_classes = Attribute('Non-Authoritative Classes', min_version='21.1')
	pkcs8_association = Attribute('PKCS8 Association')
	x509certificate_association = Attribute('X509Certificate Association')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.identity_driver import IdentityDriverAttributes


class ActiveDirectoryIdentityDriverAttributes(IdentityDriverAttributes, metaclass=IterableMeta):
	__config_class__ = "ActiveDirectory Identity Driver"
	base_dn = Attribute('Base DN')
	computer_class_name = Attribute('Computer Class Name', min_version='19.1')
	computer_query_expression = Attribute('Computer Query Expression', min_version='19.1')
	configuration = Attribute('Configuration')
	container_class_name = Attribute('Container Class Name')
	container_query_expression = Attribute('Container Query Expression')
	credential = Attribute('Credential')
	dsn = Attribute('DSN')
	distribution_list_class_name = Attribute('Distribution List Class Name')
	distribution_list_query_expression = Attribute('Distribution List Query Expression')
	friendly_name = Attribute('Friendly Name')
	group_class_name = Attribute('Group Class Name')
	group_query_expression = Attribute('Group Query Expression')
	host = Attribute('Host')
	no_rediscovery_on = Attribute('No Rediscovery On', min_version='21.2')
	port = Attribute('Port')
	resolve_nested_groups = Attribute('Resolve Nested Groups')
	search_filter = Attribute('Search Filter')
	search_root = Attribute('Search Root')
	secure = Attribute('Secure')
	timeout = Attribute('Timeout')
	user_class_name = Attribute('User Class Name')
	user_query_expression = Attribute('User Query Expression')
	vault_id = Attribute('Vault Id', min_version='18.2')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.http_ca_base import HTTPCABaseAttributes


class ThawteSPKICAAttributes(HTTPCABaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Thawte SPKI CA"
	certificate_block = Attribute('Certificate Block')
	certificate_transparency = Attribute('Certificate Transparency', min_version='16.3')
	retrieval_period = Attribute('Retrieval Period', min_version='15.4')
	server_type = Attribute('Server Type')
	signature_algorithm = Attribute('Signature Algorithm')
	uri = Attribute('URI')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.certificate_authority_base import CertificateAuthorityBaseAttributes


class ComodoCCMAttributes(CertificateAuthorityBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Comodo CCM"
	customer_login_uri = Attribute('Customer Login URI')
	organization = Attribute('Organization')
	secret_key = Attribute('Secret Key')



from pytpp.tools.logger import logger


class FeatureException(Exception):
    def log(self):
        logger.error(
            msg=self.__str__(),
            num_prev_callers=2
        )

class InvalidFormat(FeatureException): ...

class InvalidResultCode(FeatureException):
    def __init__(self, code: int, code_description: str = 'Unknown'):
        super().__init__(f'Expected a valid result code, but got "{code}": {code_description}.')

class ObjectDoesNotExist(FeatureException): ...

class UnexpectedValue(FeatureException): ...



from pytpp.api.api_base import API, APIResponse, api_response_property
from pytpp.properties.response_objects.config_schema import ConfigSchema


class _ConfigSchema:
    def __init__(self, api_obj):
        self.Attributes = self._Attributes(api_obj=api_obj)
        self.Class = self._Class(api_obj=api_obj)

    class _Attributes(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/ConfigSchema/Attributes')

        def post(self):
            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def result(self):
                    return ConfigSchema.Result(self._from_json(key='Result'))

                @property
                @api_response_property()
                def attribute_definitions(self):
                    return [ConfigSchema.AttributeDefinition(attr) for attr in self._from_json('AttributeDefinitions')]

            return _Response(response=self._post(data={}))

    class _Class(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/ConfigSchema/Class')

        def post(self, class_name: str):
            body = {
                "Class": class_name
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def result(self):
                    return ConfigSchema.Result(self._from_json(key='Result'))

                @property
                @api_response_property()
                def class_definition(self):
                    return ConfigSchema.ClassDefinition(self._from_json('ClassDefinition'))

            return _Response(response=self._post(data=body))



from pytpp.attributes._helper import IterableMeta
from pytpp.attributes.ssh_key import SSHKeyAttributes


class SSHUserKeyAttributes(SSHKeyAttributes, metaclass=IterableMeta):
	__config_class__ = "SSH User Key"



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.client_work_base import ClientWorkBaseAttributes


class ClientAgentAutomaticUpgradeWorkAttributes(ClientWorkBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Client Agent Automatic Upgrade Work"
	days_of_month = Attribute('Days Of Month')
	days_of_week = Attribute('Days Of Week')
	force_agent_upgrade = Attribute('Force Agent Upgrade')
	interval = Attribute('Interval')
	schedule_type = Attribute('Schedule Type')
	start_time = Attribute('Start Time')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.metadata_base import MetadataBaseAttributes


class MetadataTextAttributes(MetadataBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Metadata Text"
	allowed_characters = Attribute('Allowed Characters')
	error_message = Attribute('Error Message')
	mask = Attribute('Mask')
	maximum_length = Attribute('Maximum Length')
	minimum_length = Attribute('Minimum Length')
	regular_expression = Attribute('Regular Expression')



from typing import List
from pytpp.api.api_base import API, APIResponse, api_response_property
from pytpp.properties.response_objects.ssh import SSH


class _SSH:
    def __init__(self, api_obj):
        self.AddAuthorizedKey = self._AddAuthorizedKey(api_obj=api_obj)
        self.AddHostPrivateKey = self._AddHostPrivateKey(api_obj=api_obj)
        self.AddKnownHostKey = self._AddKnownHostKey(api_obj=api_obj)
        self.AddSelfServiceKey = self._AddSelfServiceKey(api_obj=api_obj)
        self.AddSelfServiceAuthorizedKey = self._AddSelfServiceAuthorizedKey(api_obj=api_obj)
        self.AddSelfServicePrivateKey = self._AddSelfServicePrivateKey(api_obj=api_obj)
        self.AddUserPrivateKey = self._AddUserPrivateKey(api_obj=api_obj)
        self.ApproveKeyOperation = self._ApproveKeyOperation(api_obj=api_obj)
        self.CancelKeyOperation = self._CancelKeyOperation(api_obj=api_obj)
        self.CancelRotation = self._CancelRotation(api_obj=api_obj)
        self.ChangePrivateKeyPassphrase = self._ChangePrivateKeyPassphrase(api_obj=api_obj)
        self.ConfirmSelfServiceKeyInstallation = self._ConfirmSelfServiceKeyInstallation(api_obj=api_obj)
        self.DeleteUnmatchedKeyset = self._DeleteUnmatchedKeyset(api_obj=api_obj)
        self.Devices = self._Devices(api_obj=api_obj)
        self.EditKeyOptions = self._EditKeyOptions(api_obj=api_obj)
        self.EditSelfServiceAuthorizedKey = self._EditSelfServiceAuthorizedKey(api_obj=api_obj)
        self.ExportSelfServiceKey = self._ExportSelfServiceKey(api_obj=api_obj)
        self.ExportSelfServicePrivateKey = self._ExportSelfServicePrivateKey(api_obj=api_obj)
        self.ImportAuthorizedKey = self._ImportAuthorizedKey(api_obj=api_obj)
        self.ImportKeyUsageData = self._ImportKeyUsageData(api_obj=api_obj)
        self.ImportPrivateKey = self._ImportPrivateKey(api_obj=api_obj)
        self.KeyDetails = self._KeyDetails(api_obj=api_obj)
        self.KeysetDetails = self._KeysetDetails(api_obj=api_obj)
        self.KeyUsage = self._KeyUsage(api_obj=api_obj)
        self.MoveKeysetsToPolicy = self._MoveKeysetsToPolicy(api_obj=api_obj)
        self.RejectKeyOperation = self._RejectKeyOperation(api_obj=api_obj)
        self.RemoveKey = self._RemoveKey(api_obj=api_obj)
        self.RetryKeyOperation = self._RetryKeyOperation(api_obj=api_obj)
        self.RetryRotation = self._RetryRotation(api_obj=api_obj)
        self.Rotate = self._Rotate(api_obj=api_obj)
        self.SetUnmatchedKeysetPassPhrase = self._SetUnmatchedKeysetPassPhrase(api_obj=api_obj)
        self.SkipKeyRotation = self._SkipKeyRotation(api_obj=api_obj)
        self.TestDeviceConnection = self._TestDeviceConnection(api_obj=api_obj)
        self.Widget = self._Widget(api_obj=api_obj)

    class _AddAuthorizedKey(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SSH/AddAuthorizedKey')

        def post(self, device_guid: str, filepath: str, keyset_id: str, username: str, allowed_source_restriction: list = None,
                 denied_source_restriction: list = None, forced_command: str = None, format: str = None, options: list = None):
            body = {
                'AllowedSourceRestriction': allowed_source_restriction,
                'DeniedSourceRestriction': denied_source_restriction,
                'DeviceGuid': device_guid,
                'Filepath': filepath,
                'ForcedCommand': forced_command,
                'Format': format,
                'KeysetId': keyset_id,
                'Options': options,
                'Username': username
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def key_id(self) -> str:
                    return self._from_json('KeyId')

                @property
                @api_response_property()
                def response(self):
                    return SSH.Response(self._from_json('Response'))

            return _Response(response=self._post(data=body))

    class _AddHostPrivateKey(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SSH/AddHostPrivateKey')

        def post(self, device_guid: str, filepath: str, username: str, format: str = None):
            body = {
                'DeviceGuid': device_guid,
                'Filepath': filepath,
                'Format': format,
                'Username': username
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def key_id(self) -> str:
                    return self._from_json('KeyId')

                @property
                @api_response_property()
                def keyset_id(self) -> str:
                    return self._from_json('KeysetId')

                @property
                @api_response_property()
                def response(self):
                    return SSH.Response(self._from_json('Response'))

            return _Response(response=self._post(data=body))

    class _AddKnownHostKey(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SSH/AddKnownHostKey')

        def post(self, device_guid: str, filepath: str, keyset_id: str, username: str, format: str = None):
            body = {
                'DeviceGuid': device_guid,
                'Filepath': filepath,
                'Format': format,
                'KeysetId': keyset_id,
                'Username': username
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def key_id(self) -> str:
                    return self._from_json('KeyId')

                @property
                @api_response_property()
                def response(self):
                    return SSH.Response(self._from_json('Response'))

            return _Response(response=self._post(data=body))

    class _AddSelfServiceKey(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SSH/AddSelfServiceKey')

        def post(self, folder_id: str, location: str, notes: str, owner: str, contact_email: str = None, keyset_id: str = None):
            body = {
                'ContactEmail': contact_email,
                'FolderId': folder_id,
                'KeysetId': keyset_id,
                'Location': location,
                'Notes': notes,
                'Owner': owner
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def key_id(self) -> str:
                    return self._from_json('KeyId')

                @property
                @api_response_property()
                def keyset_id(self) -> str:
                    return self._from_json('KeysetId')

                @property
                @api_response_property()
                def notes(self) -> str:
                    return self._from_json('Notes')

                @property
                @api_response_property()
                def response(self):
                    return SSH.Response(self._from_json('Response'))

            return _Response(response=self._post(data=body))

    class _AddSelfServiceAuthorizedKey(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SSH/AddSelfServiceAuthorizedKey')

        def post(self, allowed_source_restriction: list, denied_source_restriction: list, folder_id: str, location: str, notes: str,
                 options: list, owner: str, contact_email: str = None, forced_command: str = None, keyset_id: str = None):
            body = {
                'AllowedSourceRestriction': allowed_source_restriction,
                'ContactEmail': contact_email,
                'DeniedSourceRestriction': denied_source_restriction,
                'FolderId': folder_id,
                'ForcedCommand': forced_command,
                'KeysetId': keyset_id,
                'Location': location,
                'Notes': notes,
                'Options': options,
                'Owner': owner
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def key_id(self) -> str:
                    return self._from_json('KeyId')

                @property
                @api_response_property()
                def keyset_id(self) -> str:
                    return self._from_json('KeysetId')

                @property
                @api_response_property()
                def notes(self) -> str:
                    return self._from_json('Notes')

                @property
                @api_response_property()
                def response(self):
                    return SSH.Response(self._from_json('Response'))

            return _Response(response=self._post(data=body))

    class _AddSelfServicePrivateKey(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SSH/AddSelfServicePrivateKey')

        def post(self, folder_id: str, location: str, notes: str, owner: str, contact_email: str = None, keyset_id: str = None):
            body = {
                'ContactEmail': contact_email,
                'FolderId': folder_id,
                'KeysetId': keyset_id,
                'Location': location,
                'Notes': notes,
                'Owner': owner
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def owner(self) -> str:
                    return self._from_json('Owner')

                @property
                @api_response_property()
                def location(self) -> str:
                    return self._from_json('Location')

                @property
                @api_response_property()
                def response(self):
                    return SSH.Response(self._from_json('Response'))

            return _Response(response=self._post(data=body))

    class _AddUserPrivateKey(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SSH/AddUserPrivateKey')

        def post(self, device_guid: str, filepath: str, username: str, format: str = None, keyset_id: str = None, passphrase: str = None):
            body = {
                'DeviceGuid': device_guid,
                'Filepath': filepath,
                'Format': format,
                'KeysetId': keyset_id,
                'Passphrase': passphrase,
                'Username': username
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def key_id(self) -> str:
                    return self._from_json('KeyId')

                @property
                @api_response_property()
                def keyset_id(self) -> str:
                    return self._from_json('KeysetId')

                @property
                @api_response_property()
                def response(self):
                    return SSH.Response(self._from_json('Response'))

            return _Response(response=self._post(data=body))

    class _ApproveKeyOperation(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SSH/ApproveKeyOperation')

        def post(self, key_id: str, comment: str):
            body = {
                'KeyId': key_id,
                'Comment': comment
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def response(self):
                    return SSH.Response(self._from_json())

            return _Response(response=self._post(data=body))

    class _CancelKeyOperation(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SSH/CancelKeyOperation')

        def post(self, key_id: str):
            body = {
                'KeyId': key_id
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def response(self):
                    return SSH.Response(self._from_json())

            return _Response(response=self._post(data=body))

    class _CancelRotation(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SSH/CancelRotation')

        def post(self, keyset_id: str):
            body = {
                'KeysetId': keyset_id
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def response(self):
                    return SSH.Response(self._from_json())

            return _Response(response=self._post(data=body))

    class _ChangePrivateKeyPassphrase(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SSH/ChangePrivateKeyPassphrase')

        def post(self, key_id: str, passphrase: str):
            body = {
                'KeysetId': key_id,
                'Passphrase': passphrase
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def response(self):
                    return SSH.Response(self._from_json())

            return _Response(response=self._post(data=body))

    class _ConfirmSelfServiceKeyInstallation(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SSH/ConfirmSelfServiceKeyInstallation')

        def post(self, keyset_id: str):
            body = {
                'KeysetId': keyset_id
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def response(self):
                    return SSH.Response(self._from_json())

            return _Response(response=self._post(data=body))

    class _DeleteUnmatchedKeyset(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SSH/DeleteUnmatchedKeyset')

        def post(self, unmatched_trust_id: str):
            body = {
                'UnmatchedTrustId': unmatched_trust_id
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def response(self):
                    return SSH.Response(self._from_json())

            return _Response(response=self._post(data=body))

    class _Devices(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SSH/Devices')

        def post(self, page_size: int, offset: int = None, ssh_device_filter: dict = None):
            body = {
                'PageSize': page_size,
                'Offset': offset,
                'SshDeviceFilter': ssh_device_filter
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def data(self):
                    return [SSH.DeviceData(d) for d in self._from_json('Data')]

            return _Response(response=self._post(data=body))

    class _EditSelfServiceAuthorizedKey(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SSH/EditSelfServiceAuthorizedKey')

        def post(self, key_id: str, allowed_source_restriction: list = None, comments: str = None, denied_source_restriction: list = None,
                 forced_command: str = None, location: str = None, notes: str = None, options: list = None):
            body = {
                'AllowedSourceRestriction': allowed_source_restriction,
                'Comments': comments,
                'DeniedSourceRestriction': denied_source_restriction,
                'ForcedCommand': forced_command,
                'KeyId': key_id,
                'Location': location,
                'Notes': notes,
                'Options': options
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def response(self):
                    return SSH.Response(self._from_json('Response'))

            return _Response(response=self._post(data=body))

    class _ExportSelfServiceKey(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SSH/ExportSelfServiceKey')

        def post(self, key_id: str, format: str = None, passphrase: str = None):
            body = {
                'KeyId': key_id,
                'Format': format,
                'Passphrase': passphrase
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def key_material(self) -> str:
                    return self._from_json('KeyMaterial')

                @property
                @api_response_property()
                def response(self):
                    return SSH.Response(self._from_json('Response'))

            return _Response(response=self._post(data=body))

    class _ExportSelfServicePrivateKey(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SSH/ExportSelfServicePrivateKey')

        def post(self, key_id: str, format: str = None, passphrase: str = None):
            body = {
                'KeyId': key_id,
                'Format': format,
                'Passphrase': passphrase
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def key_id(self) -> str:
                    return self._from_json('KeyId')

                @property
                @api_response_property()
                def keyset_id(self) -> str:
                    return self._from_json('KeysetId')

                @property
                @api_response_property()
                def notes(self) -> str:
                    return self._from_json('Notes')

                @property
                @api_response_property()
                def response(self):
                    return SSH.Response(self._from_json('Response'))

            return _Response(response=self._post(data=body))

    class _ImportAuthorizedKey(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SSH/ImportAuthorizedKey')

        def post(self, device_guid: str, filepath: str, format: str, key_content_base_64: str, username: str):
            body = {
                'DeviceGuid': device_guid,
                'Filepath': filepath,
                'Format': format,
                'KeyContentBase64': key_content_base_64,
                'Username': username
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def key_id(self) -> str:
                    return self._from_json('KeyId')

                @property
                @api_response_property()
                def response(self):
                    return SSH.Response(self._from_json('SshWebResponse'))

            return _Response(response=self._post(data=body))

    class _ImportKeyUsageData(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SSH/ImportKeyUsageData')

        def post(self, log_data: list):
            body = {
                'LogData': log_data
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def response(self):
                    return SSH.Response(self._from_json('SshWebResponse'))

            return _Response(response=self._post(data=body))

    class _ImportPrivateKey(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SSH/ImportPrivateKey')

        def post(self, device_guid: str, filepath: str, format: str, key_content_base_64: str, username: str, passphrase: str = None):
            body = {
                'DeviceGuid': device_guid,
                'Filepath': filepath,
                'Format': format,
                'KeyContentBase64': key_content_base_64,
                'Passphrase': passphrase,
                'Username': username
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def key_id(self) -> str:
                    return self._from_json('KeyId')

                @property
                @api_response_property()
                def response(self):
                    return SSH.Response(self._from_json('SshWebResponse'))

            return _Response(response=self._post(data=body))

    class _EditKeyOptions(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SSH/EditKeyOptions')

        def post(self, key_id: str, allowed_source_restriction: list = None, denied_source_restriction: list = None,
                 forced_command: str = None, options: list = None):
            body = {
                'AllowedSourceRestriction': allowed_source_restriction,
                'DeniedSourceRestriction': denied_source_restriction,
                'ForcedCommand': forced_command,
                'KeyId': key_id,
                'Options': options
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def response(self):
                    return SSH.Response(self._from_json())

            return _Response(response=self._post(data=body))

    class _KeyDetails(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SSH/KeyDetails')

        def post(self, key_id: list):
            body = {
                'KeyId': key_id
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def key_data(self):
                    return [SSH.KeyData(d) for d in self._from_json('KeyData')]

            return _Response(response=self._post(data=body))

    class _KeysetDetails(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SSH/KeysetDetails')

        def get(self, keyset_id: list, load_key_data: bool = None):
            params = {
                'KeysetId': keyset_id,
                'LoadKeyData': load_key_data
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def access(self):
                    return self._from_json('Access')

                @property
                @api_response_property()
                def algorithm(self):
                    return self._from_json('Algorithm')

                @property
                @api_response_property()
                def fingerprint_md5(self):
                    return self._from_json('FingerprintMD5')

                @property
                @api_response_property()
                def fingerprint_sha256(self):
                    return self._from_json('FingerprintSHA256')

                @property
                @api_response_property()
                def keyset_id(self):
                    return self._from_json('KeysetId')

                @property
                @api_response_property()
                def last_rotation_date(self):
                    return self._from_json('LastRotationDate')

                @property
                @api_response_property()
                def last_used(self):
                    return self._from_json('LastUsed')

                @property
                @api_response_property()
                def length(self):
                    return self._from_json('Length')

                @property
                @api_response_property()
                def private_keys(self):
                    return [SSH.KeyData(key) for key in self._from_json('PrivateKeys')] if 'PrivateKeys' in self.api_response.json() else []

                @property
                @api_response_property()
                def process_status(self):
                    return self._from_json('ProcessStatus')

                @property
                @api_response_property()
                def public_keys(self):
                    return [SSH.KeyData(key) for key in self._from_json('PublicKeys')] if 'PublicKeys' in self.api_response.json() else []

                @property
                @api_response_property()
                def rotation_stage(self):
                    return self._from_json('RotationStage')

                @property
                @api_response_property()
                def type(self):
                    return self._from_json('Type')

                @property
                @api_response_property()
                def violation_status(self):
                    return self._from_json('ViolationStatus')

            return _Response(response=self._get(params=params))

        def post(self, page_size: int, keyset_filter: list = None, load_key_data: bool = None, offset: int = None):
            body = {
                'KeysetFilter': keyset_filter,
                'LoadKeyData': load_key_data,
                'Offset': offset,
                'PageSize': page_size
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def data(self):
                    return [SSH.KeySetData(key) for key in self._from_json('Data')]

            return _Response(response=self._post(data=body))

    class _KeyUsage(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SSH/KeyUsage')

        def post(self, ssh_key_usage_filter: list, load_key_data: bool = None, page_size: int = None, offset: int = None):
            body = {
                'SshKeyUsageFilter': ssh_key_usage_filter,
                'LoadKeyData': load_key_data,
                'PageSize': page_size,
                'Offset': offset
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def data(self):
                    return [SSH.KeyUsageData(key) for key in self._from_json('Data')]

            return _Response(response=self._post(data=body))

    class _MoveKeysetsToPolicy(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SSH/MoveKeysetsToPolicy')

        def post(self, keyset_ids: list, policy_path: str):
            body = {
                'KeysetIds': keyset_ids,
                'PolicyPath': policy_path
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def response(self):
                    return SSH.Response(self._from_json())

            return _Response(response=self._post(data=body))

    class _RejectKeyOperation(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SSH/RejectKeyOperation')

        def post(self, key_id: str, comment: str):
            body = {
                'KeyId': key_id,
                'Comment': comment
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def response(self):
                    return SSH.Response(self._from_json())

            return _Response(response=self._post(data=body))

    class _RemoveKey(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SSH/RemoveKey')

        def post(self, key_id: str):
            body = {
                'KeyId': key_id
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def response(self):
                    return SSH.Response(self._from_json())

            return _Response(response=self._post(data=body))

    class _RetryKeyOperation(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SSH/RetryKeyOperation')

        def post(self, key_id: str):
            body = {
                'KeyId': key_id
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def response(self):
                    return SSH.Response(self._from_json())

            return _Response(response=self._post(data=body))

    class _RetryRotation(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SSH/RetryRotation')

        def post(self, keyset_id: str):
            body = {
                'KeysetId': keyset_id
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def response(self):
                    return SSH.Response(self._from_json())

            return _Response(response=self._post(data=body))

    class _Rotate(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SSH/Rotate')

        def post(self, allow_skip_on_rotation: bool, keyset_id: str):
            body = {
                'AllowSkipOnRotation': allow_skip_on_rotation,
                'KeysetId': keyset_id
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def response(self):
                    return SSH.Response(self._from_json())

            return _Response(response=self._post(data=body))

    class _SetUnmatchedKeysetPassPhrase(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SSH/SetUnmatchedKeysetPassPhrase')

        def post(self, passphrase: str, unmatched_trust_id: str):
            body = {
                'Passphrase': passphrase,
                'UnmatchedTrustId': unmatched_trust_id
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def response(self):
                    return SSH.Response(self._from_json())

            return _Response(response=self._post(data=body))

    class _SkipKeyRotation(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SSH/SkipKeyRotation')

        def post(self, key_id: str):
            body = {
                'KeyId': key_id
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def response(self):
                    return SSH.Response(self._from_json())

            return _Response(response=self._post(data=body))

    class _TestDeviceConnection(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='SSH/TestDeviceConnection')

        def post(self, device_guids: List[str]):
            body = {
                'deviceGuids': device_guids
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def connection_results(self):
                    return [SSH.ConnectionResult(result) for result in self._from_json()]

            return _Response(response=self._post(data=body))

    class _Widget:
        def __init__(self, api_obj):
            self.CriticalAlerts = self._CriticalAlerts(api_obj=api_obj)
            self.PolicyViolations = self._PolicyViolations(api_obj=api_obj)
            self.Stats = self._Stats(api_obj=api_obj)

        class _CriticalAlerts(API):
            def __init__(self, api_obj):
                super().__init__(api_obj=api_obj, url='/SSH/Widget/CriticalAlerts')

            def get(self, min_allowed_key_length: int):
                params = {
                    'minAllowedKeyLength': min_allowed_key_length
                }

                class _Response(APIResponse):
                    def __init__(self, response):
                        super().__init__(response=response)

                    @property
                    @api_response_property()
                    def root_orphans(self) -> int:
                        return self._from_json('RootOrphans')

                    @property
                    @api_response_property()
                    def non_root_orphans(self) -> int:
                        return self._from_json('NonRootOrphans')

                    @property
                    @api_response_property()
                    def accessible_root_accounts(self) -> int:
                        return self._from_json('AccessibleRootAccounts')

                    @property
                    @api_response_property()
                    def shared_private_keys(self) -> int:
                        return self._from_json('SharedPrivateKeys')

                    @property
                    @api_response_property()
                    def non_compliant_duplicate_private_keys(self) -> int:
                        return self._from_json('NonCompliantDuplicatePrivateKeys')

                    @property
                    @api_response_property()
                    def very_small_key(self) -> int:
                        return self._from_json('VerySmallKey')

                return _Response(response=self._get(params=params))

        class _PolicyViolations(API):
            def __init__(self, api_obj):
                super().__init__(api_obj=api_obj, url='/SSH/Widget/PolicyViolations')

            def get(self):
                class _Response(APIResponse):
                    def __init__(self, response):
                        super().__init__(response=response)

                    @property
                    @api_response_property()
                    def non_compliant_force_command(self):
                        return self._from_json('NonCompliantForceCommand')

                    @property
                    @api_response_property()
                    def non_compliant_source_restriction(self):
                        return self._from_json('NonCompliantSourceRestriction')

                    @property
                    @api_response_property()
                    def missing_options(self):
                        return self._from_json('MissingOptions')

                    @property
                    @api_response_property()
                    def non_compliant_algorithm(self):
                        return self._from_json('NonCompliantAlgorithm')

                    @property
                    @api_response_property()
                    def vulnerable_protocol(self):
                        return self._from_json('VulnerableProtocol')

                    @property
                    @api_response_property()
                    def non_compliant_vendor_format(self):
                        return self._from_json('NonCompliantVendorFormat')

                    @property
                    @api_response_property()
                    def key_older_than_policy(self):
                        return self._from_json('KeyOlderThanPolicy')

                    @property
                    @api_response_property()
                    def shared_server_account(self):
                        return self._from_json('SharedServerAccount')

                    @property
                    @api_response_property()
                    def key_smaller_than_policy(self):
                        return self._from_json('KeySmallerThanPolicy')

                    @property
                    @api_response_property()
                    def duplicate_private_keys(self):
                        return self._from_json('DuplicatePrivateKeys')

                    @property
                    @api_response_property()
                    def root_access(self):
                        return self._from_json('RootAccess')

                return _Response(response=self._get())

        class _Stats(API):
            def __init__(self, api_obj):
                super().__init__(api_obj=api_obj, url='/SSH/Widget/Stats')

            def get(self, group_by: str):
                params = {
                    'GroupBy': group_by
                }

                class _Response(APIResponse):
                    def __init__(self, response):
                        super().__init__(response=response)

                    @property
                    @api_response_property()
                    def name_value_pairs(self) -> dict:
                        return self._from_json()

                return _Response(response=self._get(params=params))



from typing import Union
import re
import json
from pytpp.api.session import Session
from requests import Response, HTTPError
from pytpp.tools.logger import logger, LogTags


def api_response_property(return_on_204: type = None):
    """
    This function serves as a decorator for all API response objects. Each
    response property must be validated before returning the object. This
    function depends upon the API class.

    The ``on_204`` parameter suggests what data type is expected to be
    returned when the response status code is valid, but has no content.
    Since there is no content to return, an empty object representing that
    object will be returned instead. For example,

    .. code-block:: python

        # No 204 should ever be returned, so just validate the response
        # status codes and return the object on 200.
        @property
        @api_response_property()
        def value1(self) -> str:
            return self._from_json(key='Value1')

        # 204 could be returned by TPP, so just send an empty response
        # of the same object type as the expected response.
        @property
        @api_response_property(on_204=list)
        def value2(self) -> list:
            return self._from_json(key='Value2')

    Args:
        return_on_204: type

    Returns: Key of response content returned by TPP.

    """
    def pre_validation(func):
        def wrap(self: APIResponse, *args, **kwargs):
            if not self._validated:
                self._validate()
            if return_on_204 and self.api_response.status_code == 204:
                if callable(return_on_204):
                    return return_on_204()
                else:
                    return return_on_204
            return func(self, *args, **kwargs)
        return wrap
    return pre_validation


class API:
    """
    This is the backbone of all API definitions. It performs all requests,
    validations, logging, re-authentication, and holds the raw response. This
    class MUST be inherited by all API definitions.
    """

    def __init__(self, api_obj, url: str):
        """
        Args:
            api_obj: This is passed down from the API type object (eg. WebSDK, etc.) and
                represents that type. This type is REQUIRED because it contains the
                authenticated sessions, base URL, and re-authentication methods. It is
                through these properties this class is able to send and receive requests
                to TPP.
            url: This is the URL extension from the base URL.
        """
        self._api_obj = api_obj
        if not url.startswith('/'):
            url = '/' + url
        self._url = self._api_obj._base_url + url
        self.retries = 3

    @property
    def _session(self) -> 'Session':
        return self._api_obj._session

    def _is_api_key_invalid(self, response: Response):
        """
        Uses a regular expression to search the response text for an indication that the API key
        is expired.

        Args:
            response: The raw response object.

        Returns:
            Returns True if the API key expired. Otherwise False.

        """
        return response.status_code == 401 and bool(re.match('.*API key.*is not valid.*', response.text))

    def _delete(self, params: dict = None):
        """
        Performs a DELETE method request. If the response suggests the API Key is expired, then
        a single attempt is made to re-authenticate to TPP. Otherwise, the raw response is returned.

        Returns:
            Returns the raw JSON response.
        """
        self._log_rest_call(method='DELETE', data=params)
        retried = 0
        exc = None
        while retried < self.retries:
            try:
                response = self._session.delete(url=self._url, params=params)
                self._log_response(response=response)
                if self._is_api_key_invalid(response=response):
                    self._re_authenticate()
                    return self._session.delete(url=self._url, params=params)
                return response
            except (ConnectionResetError, ConnectionError) as e:
                exc = e
            retried += 1
        raise exc

    def _get(self, params: dict = None):
        """
        Performs a GET method request. If the response suggests the API Key is expired, then
        a single attempt is made to re-authenticate to TPP. Otherwise, the raw response is returned.

        Args:
            params: A dictionary of URL parameters to append to the URL.

        Returns:
            Returns the raw JSON response.
        """
        self._log_rest_call(method='GET', data=params)
        retried = 0
        exc = None
        while retried < self.retries:
            try:
                response = self._session.get(url=self._url, params=params)
                self._log_response(response=response)
                if self._is_api_key_invalid(response=response):
                    self._re_authenticate()
                    return self._session.get(url=self._url, params=params)
                return response
            except (ConnectionResetError, ConnectionError) as e:
                exc = e
            retried += 1
        raise exc

    def _post(self, data: Union[list, dict]):
        """
        Performs a POST method request. If the response suggests the API Key is expired, then
        a single attempt is made to re-authenticate to TPP. Otherwise, the raw response is returned.

        Args:
            data: A dictionary of data to send with the URL.

        Returns:
            Returns the raw JSON response.
        """
        self._log_rest_call(method='POST', data=data)
        retried = 0
        exc = None
        while retried < self.retries:
            try:
                response = self._session.post(url=self._url, data=data)
                self._log_response(response=response)
                if self._is_api_key_invalid(response=response):
                    self._re_authenticate()
                    return self._session.post(url=self._url, data=data)
                return response
            except (ConnectionResetError, ConnectionError) as e:
                exc = e
            retried += 1
        raise exc

    def _put(self, data: Union[list, dict]):
        """
        Performs a PUT method request. If the response suggests the API Key is expired, then
        a single attempt is made to re-authenticate to TPP. Otherwise, the raw response is returned.

        Args:
            data: A dictionary of data to send with the URL.

        Returns:
            Returns the raw JSON response.
        """
        self._log_rest_call(method='PUT', data=data)
        retried = 0
        exc = None
        while retried < self.retries:
            try:
                response = self._session.put(url=self._url, data=data)
                self._log_response(response=response)
                if self._is_api_key_invalid(response=response):
                    self._re_authenticate()
                    self._session.put(url=self._url, data=data)
                return response
            except (ConnectionResetError, ConnectionError) as e:
                exc = e
            retried += 1
        raise exc

    def _re_authenticate(self):
        """
        The current API token expired and the session needs to be re-authenticated.
        """
        logger.log(
            msg=f'{self._api_obj.__class__.__name__} API authentication token expired. Re-authenticating...',
            log_tag=LogTags.api
        )
        self._api_obj.re_authenticate()

    def _log_api_deprecated_warning(self, alternate_api: str = None, num_prev_callers=2):
        msg = f'API DEPRECATION WARNING: {self._url} is no longer supported by Venafi.'
        if alternate_api:
            msg += f'\nUse {alternate_api} instead.'
        logger.warning(
            msg=msg,
            num_prev_callers=num_prev_callers
        )

    def _log_rest_call(self, method: str, data: dict = None, num_prev_callers: int = 3):
        """
        Logs the URL and any additional data. This enforces consistency in logging across all API calls.
        """
        if data:
            payload = logger._jsonpickle.dumps(data)
            logger.log(
                msg=f'{method}\nURL: {self._url}\nBODY: {payload}',
                log_tag=LogTags.api,
                num_prev_callers=num_prev_callers
            )
        else:
            logger.log(
                msg=f'{method}\nURL: {self._url}',
                log_tag=LogTags.api,
                num_prev_callers=3
            )

    def _log_response(self, response: Response, num_prev_callers: int = 3):
        """
        Logs the URL, response code, and the content returned by TPP.
        This enforces consistency in logging across all API calls.
        """
        try:
            pretty_json = logger._jsonpickle.dumps(response.json())
        except json.JSONDecodeError:
            pretty_json = response.text or response.reason or 'No Content'
        except:
            pretty_json = 'No Content'

        logger.log(
            msg=f'URL: "{self._url}"\nRESPONSE CODE: {response.status_code}\nCONTENT: {pretty_json}',
            log_tag=LogTags.api,
            num_prev_callers=num_prev_callers
        )

class APIResponse:
    def __init__(self, response: Response):
        """
        """
        self._api_response = response
        self._decoded_api_response = None
        self._validated = False

    @property
    def api_response(self):
        return self._api_response

    @api_response.setter
    def api_response(self, value):
        # When set, a new raw response is stored and hasn't been validated, so invalidate.
        self._validated = False
        self._api_response = value

    def assert_valid_response(self):
        """
        Use this method when no response property is available after an API call or to simply
        throw an error if the return code is invalid. This simply asserts that a valid response
        status code was returned by TPP.
        """
        if not self._validated:
            self._validate()

    def is_valid_response(self):
        """
        Returns ``True`` when the response is valid, meaning a valid return code was returned by
        TPP, otherwise ``False``.
        """
        try:
            self._validate()
            return True
        except:
            return False

    def _from_json(self, key: str = None, error_key: str = None, return_on_error: type = None):
        """
        Returns the particular key within the response dictionary. If no key is provided, then
        the full response is returned as a dictionary.

        If ``key`` is provided, then it is searched in the response content. It is not case
        sensitive. If ``key`` is not found, an error is thrown and logged.

        TPP often returns a key with an error message. When an ``error_key`` is provided and an
        error message is returned an error is thrown and logged with the message provided by TPP.

        Args:
            key: Key within the top level of the response content.
            error_key: Error key within the top level of the response content.
            return_on_error: If an error occurs in retrieving content or accessing a key, this
                type will be returned instead of throwing an error.

        Returns:
            If a key is provided, returns the response content at that key. Otherwise, the full
            response content.
        """
        try:
            result = self.api_response.json()
        except json.decoder.JSONDecodeError:
            if return_on_error:
                return return_on_error()
            raise InvalidResponseError(f'{self.api_response.url} return no content. '
                                       f'Got status code {self.api_response.status_code}.')
        if error_key and error_key in result.keys():
            raise InvalidResponseError('An error occurred: "%s"' % result[error_key])
        if not key:
            return result
        for k in result.keys():
            if key.lower() == k.lower():
                return result[k]

        if return_on_error:
            return return_on_error()
        raise KeyError(f'{key} was not returned by TPP.')

    def _validate(self):
        """
        Validates the current response property by validating that there are expected return codes
        and that the actual return code is one of the valid return codes. If the return code is
        invalid, an error is thrown and logged.
        """
        self._validated = True
        try:
            self.api_response.raise_for_status()
        except HTTPError as err:
            error_msg = self.api_response.text or self.api_response.reason or 'No error message found.'
            body = logger._jsonpickle.dumps(
                json.loads(err.request.body)
            ) if err.request.body else ''
            raise InvalidResponseError('\n'.join(err.args) + f"\nERROR: {error_msg}\nCONTENT: {body}")


class InvalidResponseError(Exception):
    pass



from typing import List
from pytpp.api.api_base import API, APIResponse, api_response_property


class _Crypto:
    def __init__(self, api_obj):
        self.AvailableKeys = self._AvailableKeys(api_obj=api_obj)
        self.DefaultKey = self._DefaultKey(api_obj=api_obj)

    class _AvailableKeys(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Crypto/AvailableKeys')

        def get(self):
            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def keynames(self) -> List[str]:
                    return self._from_json('Keynames')

            return _Response(response=self._get())

    class _DefaultKey(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Crypto/DefaultKey')

        def get(self):
            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def default_key(self) -> str:
                    return self._from_json('DefaultKey')

            return _Response(response=self._get())



from pytpp.attributes._helper import IterableMeta, Attribute


class ScheduleBaseAttributes(metaclass=IterableMeta):
	__config_class__ = "Schedule Base"
	blackout = Attribute('Blackout')
	days_of_month = Attribute('Days Of Month')
	days_of_week = Attribute('Days Of Week')
	days_of_year = Attribute('Days Of Year')
	hour = Attribute('Hour')
	minute = Attribute('Minute')
	priority = Attribute('Priority')
	reschedule = Attribute('Reschedule')
	start_time = Attribute('Start Time')
	stop_time = Attribute('Stop Time')
	timezone = Attribute('Timezone', min_version='21.2')
	utc = Attribute('UTC')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.top import TopAttributes


class SSHCATemplateAttributes(TopAttributes, metaclass=IterableMeta):
	__config_class__ = "SSH CA Template"
	allow_private_key_reuse = Attribute('Allow Private Key Reuse', min_version='21.3')
	allow_to_specify_certificate_identifier_on_request = Attribute('Allow To Specify Certificate Identifier On Request', min_version='21.2')
	allow_to_specify_extensions_on_request = Attribute('Allow To Specify Extensions On Request', min_version='21.2')
	allow_to_specify_force_command_on_request = Attribute('Allow To Specify Force Command On Request', min_version='21.2')
	allow_to_specify_principals_on_request = Attribute('Allow To Specify Principals On Request', min_version='21.2')
	allow_to_specify_source_addresses_on_request = Attribute('Allow To Specify Source Addresses On Request', min_version='21.2')
	allowed_custom_extensions = Attribute('Allowed Custom Extensions', min_version='21.3')
	allowed_extension = Attribute('Allowed Extension', min_version='21.2')
	allowed_forced_command_pattern = Attribute('Allowed Forced Command Pattern', min_version='21.2')
	allowed_principal_pattern = Attribute('Allowed Principal Pattern', min_version='21.2')
	allowed_private_key_algorithm = Attribute('Allowed Private Key Algorithm', min_version='21.2')
	allowed_source_address = Attribute('Allowed Source Address', min_version='21.2')
	certificate_container = Attribute('Certificate Container', min_version='21.2')
	certificate_identifier = Attribute('Certificate Identifier', min_version='21.2')
	certificate_issuance_timeout_before_returning_pending_response = Attribute('Certificate Issuance Timeout Before Returning Pending Response', min_version='21.4')
	certificate_object_naming_pattern = Attribute('Certificate Object Naming Pattern', min_version='21.4')
	certificate_type = Attribute('Certificate Type', min_version='21.2')
	extension = Attribute('Extension', min_version='21.2')
	fingerprint_sha256 = Attribute('Fingerprint SHA256', min_version='21.2')
	force_command = Attribute('Force Command', min_version='21.2')
	hash_algorithm = Attribute('Hash Algorithm', min_version='21.2')
	include_issued_certificate_in_request = Attribute('Include Issued Certificate In Request', min_version='21.4')
	key_algorithm = Attribute('Key Algorithm', min_version='21.2')
	key_id_format = Attribute('Key ID Format', min_version='21.2')
	maximum_age = Attribute('Maximum Age', min_version='21.2')
	principal = Attribute('Principal', min_version='21.2')
	public_signing_key_vault_id = Attribute('Public Signing Key Vault Id', min_version='21.2')
	ssh_certificate_issuance_flow = Attribute('SSH Certificate Issuance Flow', min_version='21.2')
	signing_key_algorithm = Attribute('Signing Key Algorithm', min_version='21.2')
	signing_key_dn = Attribute('Signing Key DN', min_version='21.2')
	source_address = Attribute('Source Address', min_version='21.2')
	use_certificate_object_naming_pattern = Attribute('Use Certificate Object Naming Pattern', min_version='21.4')



from typing import List 
from pytpp.api.api_base import API, APIResponse, api_response_property
from pytpp.properties.response_objects.permissions import Permissions


class _Permissions:
    def __init__(self, api_obj):
        self.Object = self._Object(api_obj=api_obj)
        self.Refresh = self._Refresh(api_obj=api_obj)

    class _Object:
        def __init__(self, api_obj):
            self._api_obj = api_obj

        def Guid(self, guid):
            return self._Guid(guid=guid, api_obj=self._api_obj)

        class _Guid(API):
            def __init__(self, guid: str, api_obj):
                super().__init__(api_obj=api_obj, url=f'/Permissions/Object/{guid}')
                self._guid = guid

            def get(self):
                class _Response(APIResponse):
                    def __init__(self, response):
                        super().__init__(response=response)

                    @property
                    @api_response_property()
                    def principals(self) -> List[str]:
                        return self._from_json()

                return _Response(response=self._get())

            def Ptype(self, ptype='Local'):
                return self._Ptype(guid=self._guid, ptype=ptype, api_obj=self._api_obj)

            class _Ptype:
                def __init__(self, guid:str, ptype: str, api_obj):
                    self._guid = guid
                    self._ptype = ptype
                    self._api_obj = api_obj

                def Pname(self, pname):
                    return self._Pname(guid=self._guid, ptype=self._ptype, pname=pname, api_obj=self._api_obj)

                def Principal(self, uuid: str):
                    return self._Principal(guid=self._guid, ptype=self._ptype, uuid=uuid, api_obj=self._api_obj)

                class _Pname:
                    def __init__(self, guid: str, ptype: str, pname: str, api_obj):
                        self._guid = guid
                        self._ptype = ptype
                        self._pname = pname
                        self._api_obj = api_obj

                    def Principal(self, principal: str):
                        return self._Principal(guid=self._guid, ptype=self._ptype, pname=self._pname,
                                              principal=principal, api_obj=self._api_obj)

                    class _Principal(API):
                        def __init__(self, guid: str, ptype: str, pname: str, principal: str, api_obj):
                            super().__init__(
                                api_obj=api_obj,
                                url=f'/Permissions/Object/{guid}/{ptype}/{pname}/{principal}'
                            )
                            self.Effective = self._Effective(guid=guid, ptype=ptype, pname=pname, principal=principal, api_obj=api_obj)

                        def delete(self):
                            return APIResponse(response=self._delete())

                        def get(self):
                            class _Response(APIResponse):
                                def __init__(self, response):
                                    super().__init__(response=response)

                                @property
                                @api_response_property()
                                def explicit_permissions(self):
                                    return Permissions.Permissions(self._from_json(key='ExplicitPermissions'))

                                @property
                                @api_response_property()
                                def implicit_permissions(self):
                                    return Permissions.Permissions(self._from_json(key='ImplicitPermissions'))

                            return _Response(response=self._get())

                        def post(self, is_associate_allowed: bool = None, is_create_allowed: bool = None, is_delete_allowed: bool = None,
                                 is_manage_permissions_allowed: bool = None, is_policy_write_allowed: bool = None,
                                 is_private_key_read_allowed: bool = None, is_private_key_write_allowed: bool = None, is_read_allowed: bool = None,
                                 is_rename_allowed: bool = None, is_revoke_allowed: bool = None, is_view_allowed: bool = None,
                                 is_write_allowed: bool = None):
                            body = {
                                'IsAssociateAllowed': is_associate_allowed,
                                'IsCreateAllowed': is_create_allowed,
                                'IsDeleteAllowed': is_delete_allowed,
                                'IsManagePermissionsAllowed': is_manage_permissions_allowed,
                                'IsPolicyWriteAllowed': is_policy_write_allowed,
                                'IsPrivateKeyReadAllowed': is_private_key_read_allowed,
                                'IsPrivateKeyWriteAllowed': is_private_key_write_allowed,
                                'IsReadAllowed': is_read_allowed,
                                'IsRenameAllowed': is_rename_allowed,
                                'IsRevokeAllowed': is_revoke_allowed,
                                'IsViewAllowed': is_view_allowed,
                                'IsWriteAllowed': is_write_allowed
                            }

                            return APIResponse(response=self._post(data=body))

                        def put(self, is_associate_allowed: bool = None, is_create_allowed: bool = None, is_delete_allowed: bool = None,
                                is_manage_permissions_allowed: bool = None, is_policy_write_allowed: bool = None,
                                is_private_key_read_allowed: bool = None, is_private_key_write_allowed: bool = None, is_read_allowed: bool = None,
                                is_rename_allowed: bool = None, is_revoke_allowed: bool = None, is_view_allowed: bool = None,
                                is_write_allowed: bool = None):
                            body = {
                                'IsAssociateAllowed': is_associate_allowed,
                                'IsCreateAllowed': is_create_allowed,
                                'IsDeleteAllowed': is_delete_allowed,
                                'IsManagePermissionsAllowed': is_manage_permissions_allowed,
                                'IsPolicyWriteAllowed': is_policy_write_allowed,
                                'IsPrivateKeyReadAllowed': is_private_key_read_allowed,
                                'IsPrivateKeyWriteAllowed': is_private_key_write_allowed,
                                'IsReadAllowed': is_read_allowed,
                                'IsRenameAllowed': is_rename_allowed,
                                'IsRevokeAllowed': is_revoke_allowed,
                                'IsViewAllowed': is_view_allowed,
                                'IsWriteAllowed': is_write_allowed
                            }

                            return APIResponse(response=self._put(data=body))

                        class _Effective(API):
                            def __init__(self, guid: str, ptype: str, pname: str, principal: str, api_obj):
                                super().__init__(
                                    api_obj=api_obj,
                                    url=f'/Permissions/Object/{guid}/{ptype}/{pname}/{principal}/Effective'
                                )

                            def get(self):
                                class _Response(APIResponse):
                                    def __init__(self, response):
                                        super().__init__(response=response)

                                    @property
                                    @api_response_property()
                                    def effective_permissions(self):
                                        return Permissions.Permissions(self._from_json('EffectivePermissions'))

                                return _Response(response=self._get())

                class _Principal(API):
                    def __init__(self, guid: str, ptype: str, uuid: str, api_obj):
                        super().__init__(
                            api_obj=api_obj,
                            url=f'/Permissions/Object/{guid}/{ptype}/{uuid}'
                        )
                        self.Effective = self._Effective(guid=guid, uuid=uuid, api_obj=api_obj)

                    def delete(self):
                        return APIResponse(response=self._delete())

                    def get(self):
                        class _Response(APIResponse):
                            def __init__(self, response):
                                super().__init__(response=response)

                            @property
                            @api_response_property()
                            def explicit_permissions(self):
                                return Permissions.Permissions(self._from_json('ExplicitPermissions'))

                            @property
                            @api_response_property()
                            def implicit_permissions(self):
                                return Permissions.Permissions(self._from_json('ImplicitPermissions'))

                        return _Response(response=self._get())

                    def post(self, is_associate_allowed: bool = None, is_create_allowed: bool = None, is_delete_allowed: bool = None,
                             is_manage_permissions_allowed: bool = None, is_policy_write_allowed: bool = None,
                             is_private_key_read_allowed: bool = None, is_private_key_write_allowed: bool = None, is_read_allowed: bool = None,
                             is_rename_allowed: bool = None, is_revoke_allowed: bool = None, is_view_allowed: bool = None,
                             is_write_allowed: bool = None):
                        body = {
                           'IsAssociateAllowed': is_associate_allowed,
                           'IsCreateAllowed': is_create_allowed,
                           'IsDeleteAllowed': is_delete_allowed,
                           'IsManagePermissionsAllowed': is_manage_permissions_allowed,
                           'IsPolicyWriteAllowed': is_policy_write_allowed,
                           'IsPrivateKeyReadAllowed': is_private_key_read_allowed,
                           'IsPrivateKeyWriteAllowed': is_private_key_write_allowed,
                           'IsReadAllowed': is_read_allowed,
                           'IsRenameAllowed': is_rename_allowed,
                           'IsRevokeAllowed': is_revoke_allowed,
                           'IsViewAllowed': is_view_allowed,
                           'IsWriteAllowed': is_write_allowed
                        }

                        return APIResponse(response=self._post(data=body))

                    def put(self, is_associate_allowed: bool = None, is_create_allowed: bool = None, is_delete_allowed: bool = None,
                            is_manage_permissions_allowed: bool = None, is_policy_write_allowed: bool = None,
                            is_private_key_read_allowed: bool = None, is_private_key_write_allowed: bool = None, is_read_allowed: bool = None,
                            is_rename_allowed: bool = None, is_revoke_allowed: bool = None, is_view_allowed: bool = None,
                            is_write_allowed: bool = None):
                        body = {
                            'IsAssociateAllowed': is_associate_allowed,
                            'IsCreateAllowed': is_create_allowed,
                            'IsDeleteAllowed': is_delete_allowed,
                            'IsManagePermissionsAllowed': is_manage_permissions_allowed,
                            'IsPolicyWriteAllowed': is_policy_write_allowed,
                            'IsPrivateKeyReadAllowed': is_private_key_read_allowed,
                            'IsPrivateKeyWriteAllowed': is_private_key_write_allowed,
                            'IsReadAllowed': is_read_allowed,
                            'IsRenameAllowed': is_rename_allowed,
                            'IsRevokeAllowed': is_revoke_allowed,
                            'IsViewAllowed': is_view_allowed,
                            'IsWriteAllowed': is_write_allowed
                        }

                        return APIResponse(response=self._put(data=body))

                    class _Effective(API):
                        def __init__(self, guid: str, uuid: str, api_obj):
                            super().__init__(
                                api_obj=api_obj,
                                url=f'/Permissions/Object/{guid}/Local/{uuid}/Effective'
                            )

                        def get(self):
                            class _Response(APIResponse):
                                def __init__(self, response):
                                    super().__init__(response=response)

                                @property
                                @api_response_property()
                                def effective_permissions(self):
                                    return Permissions.Permissions(self._from_json('EffectivePermissions'))

                            return _Response(response=self._get())

    class _Refresh(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Permissions/Refresh')

        def get(self):
            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def result(self) -> int:
                    return self._from_json('Result')

            return _Response(response=self._get())



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.certificate_trust_store_base import CertificateTrustStoreBaseAttributes


class PaloAltoNetworkFWTrustStoreAttributes(CertificateTrustStoreBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Palo Alto Network FW Trust Store"
	lock_config = Attribute('Lock Config', min_version='15.1')
